import unittest
from unittest.mock import patch
import pandas as pd
from pyloghub.geocoding import forward_geocoding  # Replace 'your_module' with the actual name of your module

class TestForwardGeocoding(unittest.TestCase):
    def setUp(self):
        # Sample input data
        self.data = {
            'country': ['DE', 'DE'],
            'state': ['BW', 'BW'],
            'postalCode': ['69117', '69123'],
            'city': ['Heidelberg', 'Heidelberg'],
            'street': ['Schlosshof 1', 'Wieblinger Weg 94']
        }
        self.addresses_df = pd.DataFrame(self.data)

        # Sample output data
        self.expected_output = pd.DataFrame(
            [
                {
                    "country": "DE",
                    "state": "BW",
                    "postalCode": "69117",
                    "city": "Heidelberg",
                    "street": "Schlosshof 1",
                    "parsedCountry": "DE",
                    "parsedState": "Baden-Württemberg",
                    "parsedPostalCode": "69117",
                    "parsedCity": "Heidelberg",
                    "parsedStreet": " 1",
                    "parsedLatitude": 49.4103591,
                    "parsedLongitude": 8.7157387,
                    "validationQuality": 81
                },
                {
                    "country": "DE",
                    "state": "BW",
                    "postalCode": "69123",
                    "city": "Heidelberg",
                    "street": "Wieblinger Weg 94",
                    "parsedCountry": "DE",
                    "parsedState": "Baden-Württemberg",
                    "parsedPostalCode": "69123",
                    "parsedCity": "Heidelberg",
                    "parsedStreet": "Wieblinger Weg 94 ",
                    "parsedLatitude": 49.411858,
                    "parsedLongitude": 8.6477021,
                    "validationQuality": 86
                }
            ]
        )

    @patch('log_hub.geocoding.requests.post')
    def test_forward_geocoding(self, mock_post):
        # Mocking the API response
        mock_response = mock_post.return_value
        mock_response.status_code = 200
        mock_response.json.return_value = {
            "geocodes": self.expected_output.to_dict(orient='records')
        }

        # Call the function
        result = forward_geocoding(self.addresses_df, 'dummy_api_key')

        # Check if the result matches the expected output
        pd.testing.assert_frame_equal(result, self.expected_output)

if __name__ == '__main__':
    unittest.main()




import os
import pandas as pd
import warnings
from typing import Optional, Dict, Tuple
import logging
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import convert_timestamps, validate_and_convert_data_types, convert_df_to_dict_excluding_nan, convert_to_float
from sending_requests import post_method, create_headers, create_url, get_workspace_entities

def forward_milkrun_optimization_plus(depots: pd.DataFrame, vehicles: pd.DataFrame, jobs: pd.DataFrame, timeWindowProfiles: pd.DataFrame, breaks: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
    """
    Perform milkrun optimization based on depots, vehicles, jobs, time window profiles, and breaks.

    This function takes five DataFrames representing depots, vehicles, jobs, time window profiles, and breaks, 
    along with a parameters dictionary and an API key. It performs milkrun optimization using the Log-hub service.

    Parameters:
    depots (pd.DataFrame): DataFrame containing depot information.
        Columns:
        - country (str): Country code.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        - depotId (str): Depot Id.
        - processingTimeAtTheDepo (number): Loading or unloading time at depot.
        - timeWindowProfile (str): The ID of the time window profile for the depot that determines the working time of that depot.

    vehicles (pd.DataFrame): DataFrame containing vehicle information.
        Columns:
        - vehicleTypeId (str): Vehicle Type Id.
        - availableVehicles (number): Available Vehicles.
        - startDepot (str): Start Depot.
        - endDepot (str): End Depot.
        - maxWeight (number): Maximum Weight.
        - maxVolume (number): Maximum Volume.
        - maxPallets (number): Maximum Pallets.
        - maxStops (number): Max Stops.
        - timeWindowStart (str): Time Window Start (ISO 8601 format).
        - timeWindowEnd (str): Time Window End (ISO 8601 format).
        - profile (str): Profile.
        - speedFactor (number): Speed Factor.
        - fixed (number): Fixed cost.
        - perHour (number): Cost per hour.
        - perKilometer (number): Cost per kilometer.
        - costPerStop (number): Cost per stop.
        - minimumTravelTime (number): Min travel time.
        - maxTravelTime (number): Max Travel Time.
        - max_distance (number): Maximum number of kilometers that vehicle can drive per route.
        - maximumDistanceBetweenStops (number): Maximum number of kilometers that vehicle can drive between each stop.
        - breakId (str): Break Id.

    jobs (pd.DataFrame): DataFrame containing job information.
        Columns:
        - country (str): Country code.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        - orderId (str): Order Id.
        - weight (number): Weight.
        - volume (number): Volume.
        - pallets (number): Pallets.
        - pickupDelivery (str): Pickup/Delivery.
        - vehicleTypeId (str): Vehicle Type Id.
        - stopDuration (number): Stop Duration.
        - timeWindowProfile (str): Time Window Profile.
        - stopDurationAtDepo (number): Loading/Unloading Time at Depot.
        - external_costs (number): External costs for order.

    timeWindowProfiles (pd.DataFrame): DataFrame containing time window profile information.
        Columns:
        - timeWindowProfileId (str): Time Window Profile Id.
        - timeWindowProfileStart (str): Time Window Profile Start (ISO 8601 format).
        - timeWindowProfileEnd (str): Time Window Profile End (ISO 8601 format).

    breaks (pd.DataFrame): DataFrame containing break information.
        Columns:
        - breakId (str): Break Id.
        - earliestBreakStart (str): Earliest Break Start (ISO 8601 format).
        - latestBreakStart (str): Latest Break Start (ISO 8601 format).
        - earliestRelativeBreakStart (str): Minimum driving time before the break can start (ISO 8601 format).
        - latestRelativeBreakTime (str): Maximum driving time until the break must start (ISO 8601 format).
        - breakDuration (number): Break Duration.

    parameters (Dict): Dictionary containing parameters:
        - durationUnit: enum "min", "sec" or "h"
        - costAllocationBy: emun "allWVP", "weight", "volume" or "pallets"

    api_key (str): Log-hub API key for accessing the milkrun optimization service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]: Three pandas DataFrames containing route overview, route details, 
                                                     and external orders. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    # Convert datetime columns in each DataFrame to string format (ISO 8601)
    depots = convert_timestamps(depots)
    vehicles = convert_timestamps(vehicles)
    jobs = convert_timestamps(jobs)
    timeWindowProfiles = convert_timestamps(timeWindowProfiles)
    breaks = convert_timestamps(breaks)

    # Define expected columns and data types for each DataFrame
    depot_mandatory_columns = {
        'country': 'str', 'depotId': 'str'
    }
    depot_optional_columns = {
        'state': 'str', 'postalCode': 'str', 'city': 'str', 
        'street': 'str', 'timeWindowProfile': 'str'
    }
    depot_optional_float = ['processingTimeAtTheDepo']

    vehicle_mandatory_columns = {
        'vehicleTypeId': 'str', 'availableVehicles': 'int'
    }   
    vehicle_optional_columns = {
        'startDepot': 'str', 'endDepot': 'str', 'timeWindowStart': 'str', 'timeWindowEnd': 'str', 'profile': 'str', 'breakId': 'str'
    }
    vehicle_optional_float = ['maxWeight', 'maxVolume', 'maxPallets', 'maxStops', 'speedFactor', 'fixed', 'perHour', 'perKilometer', 'costPerStop', 'minimumTravelTime', 'maxTravelTime', 'max_distance', 'maximumDistanceBetweenStops']
    job_mandatory_columns = {
        'country': 'str', 'orderId': 'str', 'pickupDelivery': 'str', 'depotId': 'str'
    }
    job_optional_columns = {
        'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str', 'vehicleTypeId': 'str', 'timeWindowProfile': 'str'
    }
    job_optional_float = ['weight', 'volume', 'pallets', 'stopDuration', 'external_costs']
    timeWindowProfile_optional_columns = {
        'timeWindowProfileId': 'str', 'timeWindowProfileStart': 'str', 'timeWindowProfileEnd': 'str'
    } 
    break_optional_columns = {
        'breakId': 'str', 'earliestBreakStart': 'str', 'latestBreakStart': 'str', 'earliestRelativeBreakStart':'str', 'latestRelativeBreakStart': 'str'
    }
    break_optional_float = ['breakDuration']

    # Perform validation and conversion for each DataFrame
    depots = validate_and_convert_data_types(depots, depot_mandatory_columns, 'mandatory', 'depots')
    if not depots is None:
        depots = convert_to_float(depots, depot_optional_float, 'optional')
        depots = validate_and_convert_data_types(depots, depot_optional_columns, 'optional', 'depots')
        if not depots is None:
            depots = convert_df_to_dict_excluding_nan(depots, depot_optional_float)
    vehicles = validate_and_convert_data_types(vehicles, vehicle_mandatory_columns, 'mandatory', 'vehicles')
    if not vehicles is None:
        vehicles = convert_to_float(vehicles, vehicle_optional_float, 'optional')
        vehicles = validate_and_convert_data_types(vehicles, vehicle_optional_columns, 'optional', 'vehicles')
        if not vehicles is None:
            vehicles = convert_df_to_dict_excluding_nan(vehicles, vehicle_optional_float)
    jobs = validate_and_convert_data_types(jobs, job_mandatory_columns, 'mandatory', 'jobs')
    if not jobs is None:
        jobs = convert_to_float(jobs, job_optional_float, 'optional')
        jobs = validate_and_convert_data_types(jobs, job_optional_columns, 'optional', 'jobs')
        if not jobs is None:
            jobs = convert_df_to_dict_excluding_nan(jobs, job_optional_float)
    timeWindowProfiles = validate_and_convert_data_types(timeWindowProfiles, timeWindowProfile_optional_columns, 'optional', 'time window profiles')
    breaks = validate_and_convert_data_types(breaks, break_optional_columns, 'optional', 'breaks')
    if not breaks is None:
        breaks = convert_to_float(breaks, break_optional_float, 'optional')
        breaks = convert_df_to_dict_excluding_nan(breaks, break_optional_float)

    # Exit if any DataFrame validation failed
    if any(df is None for df in [depots, vehicles, jobs, timeWindowProfiles, breaks]):
        return None

    url = create_url("milkrunoptimizationplus")
    
    headers = create_headers(api_key)

    payload = {
        "depots": depots,
        "vehicles": vehicles,
        "jobs": jobs,
        "timeWindowProfiles": timeWindowProfiles.to_dict(orient='records'),
        "breaks": breaks,
        "parameters": parameters
    }
    payload = save_scenario_check(save_scenario, payload)
    response_data = post_method(url, payload, headers, "milkrun optimization plus")
    if response_data is None:
        return None
    else:
        route_overview_df = pd.DataFrame(response_data['routeOverview'])
        route_details_df = pd.DataFrame(response_data['routeDetails'])
        external_orders_df = pd.DataFrame(response_data['externalOrders'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return route_overview_df, route_details_df, external_orders_df

def forward_milkrun_optimization_plus_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'MilkrunPlusSampleDataAddresses.xlsx')
    depots_df = pd.read_excel(data_path, sheet_name='depots', usecols='A:I', dtype={'postalCode': str})
    vehicles_df = pd.read_excel(data_path, sheet_name='vehicles', usecols='A:V', dtype={'maxTravelTime': int})
    jobs_df = pd.read_excel(data_path, sheet_name='jobs', usecols='A:Q', dtype={'postalCode': str})
    time_window_profiles_df = pd.read_excel(data_path, sheet_name='timeWindowProfiles', usecols='A:D')
    breaks_df = pd.read_excel(data_path, sheet_name='breaks', usecols='A:G')

    parameters = {
        "durationUnit": "min",
        "costAllocationBy": "allWVP"
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'depots': depots_df, 'vehicles': vehicles_df, 'jobs': jobs_df, 'timeWindowProfiles': time_window_profiles_df, 'breaks': breaks_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


def reverse_milkrun_optimization_plus(depots: pd.DataFrame, vehicles: pd.DataFrame, jobs: pd.DataFrame, timeWindowProfiles: pd.DataFrame, breaks: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
    """
    Perform reverse milkrun optimization based on depots, vehicles, jobs, time window profiles, and breaks.

    This function takes five DataFrames representing depots, vehicles, jobs, time window profiles, and breaks, 
    along with a parameters dictionary and an API key. It performs reverse milkrun optimization using the Log-hub service.

    Parameters:
    depots (pd.DataFrame): DataFrame containing depot information, with columns:
        - depotId (str): Depot Id.
        - latitude (float): Latitude of the depot.
        - longitude (float): Longitude of the depot.
        - processingTimeAtTheDepo (number): Loading or unloading time at depot.
        - timeWindowProfile (str): The ID of the time window profile for the depot that determines the working time of that depot.

    vehicles (pd.DataFrame): DataFrame containing vehicle information, with columns:
        - vehicleTypeId (str): Vehicle Type Id.
        - availableVehicles (int): Available Vehicles.
        - startDepot (str): Start Depot.
        - endDepot (str): End Depot.
        - maxWeight (float): Maximum Weight.
        - maxVolume (float): Maximum Volume.
        - maxPallets (int): Maximum Pallets.
        - maxStops (int): Max Stops.
        - timeWindowStart (str): Time Window Start (ISO 8601 format).
        - timeWindowEnd (str): Time Window End (ISO 8601 format).
        - profile (str): Profile.
        - speedFactor (float): Speed Factor.
        - fixed (float): Fixed cost.
        - perHour (float): Cost per hour.
        - perKilometer (float): Cost per kilometer.
        - costPerStop (float): Cost per stop.
        - minimumTravelTime (float): Min travel time.
        - maxTravelTime (float): Max Travel Time.
        - max_distance (float): Maximum distance that vehicle can drive per route.
        - maximumDistanceBetweenStops (float): Maximum distance that vehicle can drive between two stops.
        - breakId (str): Break Id.

    jobs (pd.DataFrame): DataFrame containing job information, with columns:
        - orderId (str): Order Id.
        - latitude (float): Latitude of the job location.
        - longitude (float): Longitude of the job location.
        - depotId (str): Depot Id.
        - weight (float): Weight.
        - volume (float): Volume.
        - pallets (int): Pallets.
        - pickupDelivery (str): Pickup/Delivery.
        - depotId (str): Depot Id.
        - vehicleTypeId (str): Vehicle Type Id.
        - stopDuration (float): Stop Duration.
        - timeWindowProfile (str): Time Window Profile.
        - stopDurationAtDepo (float): Loading/Unloading Time at Depot.
        - external_costs (float): External costs.

    timeWindowProfiles (pd.DataFrame): DataFrame containing time window profile information, with columns:
        - timeWindowProfileId (str): Time Window Profile Id.
        - timeWindowProfileStart (str): Time Window Profile Start (ISO 8601 format).
        - timeWindowProfileEnd (str): Time Window Profile End (ISO 8601 format).

    breaks (pd.DataFrame): DataFrame containing break information, with columns:
        - breakId (str): Break Id.
        - earliestBreakStart (str): Earliest Break Start (ISO 8601 format).
        - latestBreakStart (str): Latest Break Start (ISO 8601 format).
        - earliestRelativeBreakStart (str): Earliest Relative Break Start (ISO 8601 format).
        - latestRelativeBreakStart (str): Latest Relative Break Start (ISO 8601 format).
        - breakDuration (float): Break Duration.

    parameters (Dict): Dictionary containing parameters:
        - durationUnit: enum "min", "sec" or "h"
        - costAllocationBy: emun "allWVP", "weight", "volume" or "pallets"

    api_key (str): Log-hub API key for accessing the reverse milkrun optimization service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]: Three pandas DataFrames containing route overview, route details, 
                                                     and external orders. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    # Convert datetime columns in each DataFrame to string format (ISO 8601)
    depots = convert_timestamps(depots)
    vehicles = convert_timestamps(vehicles)
    jobs = convert_timestamps(jobs)
    timeWindowProfiles = convert_timestamps(timeWindowProfiles)
    breaks = convert_timestamps(breaks)

    # Define expected columns and data types for each DataFrame
    depot_mandatory_columns = {
        'latitude': 'float', 'longitude': 'float', 'depotId': 'str'
    }
    depot_optional_columns = {'timeWindowProfile': 'str'
    }
    depot_optional_float = ['processingTimeAtTheDepo']
    vehicle_mandatory_columns = {
        'vehicleTypeId': 'str', 'availableVehicles': 'int'
    }
    vehicle_optional_columns = {'startDepot': 'str', 'endDepot': 'str', 'timeWindowStart': 'str', 'timeWindowEnd': 'str', 'profile': 'str', 'breakId': 'str'}
    vehicle_optional_float = ['maxWeight', 'maxVolume', 'maxPallets', 'maxStops', 'speedFactor', 'fixed', 'perHour', 'perKilometer', 'costPerStop', 'minimumTravelTime', 'maxTravelTime', 'max_distance', 'maximumDistanceBetweenStops']
    job_mandatory_columns = {
        'latitude': 'float', 'longitude': 'float', 'depotId': 'str', 'orderId': 'str', 'pickupDelivery': 'str'
    }
    job_optional_columns = {
        'vehicleTypeId': 'str', 'timeWindowProfile': 'str', 
        }
    job_optional_float = ['weight', 'volume', 'pallets', 'stopDuration', 'stopDurationAtDepo', 'external_costs']
    timeWindowProfile_optional_columns = {
        'timeWindowProfileId': 'str', 'timeWindowProfileStart': 'str', 'timeWindowProfileEnd': 'str'
    }
    break_optional_columns = {
        'breakId': 'str', 'earliestBreakStart': 'str', 'latestBreakStart': 'str', 'earliestRelativeBreakStart': 'str', 'latestRelativeBreakStart': 'str'
    }
    break_optional_float = ['breakDuration']

    # Perform validation and conversion for each DataFrame
    depots = validate_and_convert_data_types(depots, depot_mandatory_columns, 'mandatory', 'depots')
    if not depots is None:
        depots = convert_to_float(depots, depot_optional_float, 'optional')
        depots = validate_and_convert_data_types(depots, depot_optional_columns, 'optional', 'depots')
        if not depots is None:
            depots = convert_df_to_dict_excluding_nan(depots, depot_optional_float)
    vehicles = validate_and_convert_data_types(vehicles, vehicle_mandatory_columns, 'mandatory', 'vehicles')
    if not vehicles is None:
        vehicles = convert_to_float(vehicles, vehicle_optional_float, 'optional')
        vehicles = validate_and_convert_data_types(vehicles, vehicle_optional_columns, 'optional', 'vehicles')
        if not vehicles is None:
            vehicles = convert_df_to_dict_excluding_nan(vehicles, vehicle_optional_float)
    jobs = validate_and_convert_data_types(jobs, job_mandatory_columns, 'mandatory', 'jobs')
    if not jobs is None:
        jobs = convert_to_float(jobs, job_optional_float, 'optional')
        jobs = validate_and_convert_data_types(jobs, job_optional_columns, 'optional', 'jobs')
        if not jobs is None:
            jobs = convert_df_to_dict_excluding_nan(jobs, job_optional_float)
    timeWindowProfiles = validate_and_convert_data_types(timeWindowProfiles, timeWindowProfile_optional_columns, 'optional', 'time window profiles')
    breaks = validate_and_convert_data_types(breaks, break_optional_columns, 'optional', 'breaks')
    if not breaks is None:
        breaks = convert_to_float(breaks, break_optional_float, 'optional')
        breaks = convert_df_to_dict_excluding_nan(breaks, break_optional_float)

    # Exit if any DataFrame validation failed
    if any(df is None for df in [depots, vehicles, jobs, timeWindowProfiles, breaks]):
        return None

    url = create_url("reversemilkrunoptimizationplus")
    
    headers = create_headers(api_key)

    payload = {
        "depots": depots,
        "vehicles": vehicles,
        "jobs": jobs,
        "timeWindowProfiles": timeWindowProfiles.to_dict(orient='records'),
        "breaks": breaks,
        "parameters": parameters
    }
    payload = save_scenario_check(save_scenario, payload)
    response_data = post_method(url, payload, headers, "reverse milkrun optimization plus")
    if response_data is None:
        return None
    else:
        route_overview_df = pd.DataFrame(response_data['routeOverview'])
        route_details_df = pd.DataFrame(response_data['routeDetails'])
        external_orders_df = pd.DataFrame(response_data['externalOrders'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return route_overview_df, route_details_df, external_orders_df

def reverse_milkrun_optimization_plus_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'MilkrunPlusSampleDataReverse.xlsx')
    depots_df = pd.read_excel(data_path, sheet_name='depots', usecols='A:F', dtype={'postalCode': str})
    vehicles_df = pd.read_excel(data_path, sheet_name='vehicles', usecols='A:V', dtype={'maxTravelTime': int})
    jobs_df = pd.read_excel(data_path, sheet_name='jobs', usecols='A:N', dtype={'postalCode': str})
    time_window_profiles_df = pd.read_excel(data_path, sheet_name='timeWindowProfiles', usecols='A:D')
    breaks_df = pd.read_excel(data_path, sheet_name='breaks', usecols='A:G')

    parameters = {
        "durationUnit": "min",
        "costAllocationBy": "allWVP" 
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'depots': depots_df, 'vehicles': vehicles_df, 'jobs': jobs_df, 'timeWindowProfiles': time_window_profiles_df, 'breaks': breaks_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


import logging
import pandas as pd

def validate_and_convert_data_types(df, required_columns, column_type, df_name):
    """
    Validate and convert the data types of the DataFrame columns.
    Log an error message if a mandatory column is missing or if conversion fails.
    """
    for col, dtype in required_columns.items():
        if column_type == 'mandatory':
            if col not in df.columns:
                logging.error(f"In the table '{df_name}' is missing a mandatory column '{col}'")
                return None
            try:
                df[col] = df[col].astype(dtype)
            except Exception as e:
                logging.error(f"Data type conversion failed in the table {df_name} for column '{col}': {e}")
                return None
        elif column_type == 'optional':
            if col in df.columns:
                try:
                    df[col] = df[col].astype(dtype)
                except Exception as e:
                    logging.error(f"Data type conversion failed in the table {df_name} for column '{col}': {e}")
                    return None
    return df

def convert_dates(df, date_columns):
        """
        Converts the DataFrame date columns.
        """
        for col in date_columns:
            df[col] = pd.to_datetime(df[col]).dt.strftime('%Y-%m-%d')
        return df

def convert_to_string(df, string_columns):
        """
        Converts the DataFrame string columns.
        """
        for col in string_columns:
            df[col] = df[col].astype(str)
        return df

def convert_to_float(df, float_columns, column_type):
        """
        Converts the DataFrame numeric columns.
        """
        for col in float_columns:
            if column_type == 'mandatory':
                df[col] = pd.to_numeric(df[col], errors='coerce')
            elif column_type == 'optional':
                 if col in df.columns:
                      df[col] = pd.to_numeric(df[col], errors='coerce')
        return df

def convert_df_to_dict_excluding_nan(df, columns_to_check):
        """
        Convert a DataFrame to a list of dictionaries, excluding specified keys if their values are NaN.

        Parameters:
        df (pd.DataFrame): The DataFrame to convert.
        columns_to_check (list): List of column names to check for NaN values.

        Returns:
        list: A list of dictionaries representing the rows of the DataFrame, excluding keys for NaN values in specified columns.
        """
        records = []
        for _, row in df.iterrows():
            record = {}
            for column, value in row.items():
                if pd.notna(value) or column not in columns_to_check:
                    record[column] = value
            records.append(record)
        return records

def convert_timestamps(df):
        """
        Converts the DataFrame date columns to a timestamp format.
        """
        for col in df.columns:
            if pd.api.types.is_datetime64_any_dtype(df[col]):
                df[col] = df[col].dt.strftime('%Y-%m-%dT%H:%M:%S.%fZ')
        return df

def validate_boolean(value):
        """
        Validating the boolean values.
        """
        return isinstance(value, bool)


def exclude_nan_depending_on_dtype(df, columns_to_check, df_name):
    """
    Convert a DataFrame to a list of dictionaries, excluding specified keys if their values are NaN.

    Parameters:
    df (pd.DataFrame): The DataFrame to convert.
    columns_to_check (list): List of column names to check for NaN values.

    Returns:
    list: A list of dictionaries representing the rows of the DataFrame, excluding keys for NaN values in specified columns.
    """
    records = []
    if len(set(columns_to_check) - set(df.columns))>0:
        missing_columns = set(columns_to_check) - set(df.columns) 
        logging.error(f"Table {df_name} is missing mandatory column: {', '.join(list(missing_columns))}")
        return None
    else:
        for ind, row in df.iterrows():
            record = {}
            for column, value in row.items():
                if columns_to_check[column] == 'float':
                    df.loc[ind, column] = pd.to_numeric(df[column][ind], errors='coerce')
                    if pd.notna(df[column][ind]):
                        record[column] = float(df[column][ind])
                elif columns_to_check[column] == 'str':
                    df.loc[ind, column] = str(df[column][ind])
                    record[column] = df[column].loc[ind]
            records.append(record)
        return records

def remove_nonexisting_optional_columns(optional_columns, sent_columns):
    updated_optional_columns = optional_columns.copy()
    for col in optional_columns.keys():
        if not col in sent_columns:
            updated_optional_columns.pop(col) 

    return updated_optional_columns


import os
import pandas as pd
from typing import Optional
import warnings
import logging
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check,  create_button
from input_data_validation import convert_df_to_dict_excluding_nan, convert_to_float, validate_and_convert_data_types
from sending_requests import post_method, create_headers, create_url, get_workspace_entities

def forward_freight_matrix_plus(shipments_df: pd.DataFrame, matrix_id: str, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[pd.DataFrame]:
    """
    Calculate the freight matrix for a list of shipments provided in a pandas DataFrame.

    Parameters:
    shipments_df (pd.DataFrame): A pandas DataFrame containing the shipment details. Each row represents a unique shipment, and the columns include:

    - shipmentId (str): A unique identifier for the shipment. 
    - shipmentDate (str): The scheduled date and time of the shipment in UTC, in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).
    - fromLocationId (str): Identifier for the origin location.
    - fromCountry (str): ISO country code for the origin location.
    - fromState (str): State or region code for the origin location. 
    - fromCity (str): City name of the origin location. 
    - fromPostalCode (str): Postal code of the origin location. Increases address accuracy.
    - fromStreet (str): Street address of the origin location. Increases address accuracy.
    - fromZone (str): A logistics zone or area at the origin. Used for categorizing regions for shipping purposes, which can impact cost.
    - toLocationId (str): Identifier for the destination location. 
    - toCountry (str): ISO country code for the destination location. 
    - toState (str): State or region code for the destination location. 
    - toCity (str): City name of the destination location. 
    - toPostalCode (str): Postal code of the destination location. Increases address accuracy.
    - toStreet (str): Street address of the destination location. Increases address accuracy.
    - toZone (str): A logistics zone or area at the destination. Used to categorize regions for shipping purposes, impacting delivery strategies and costs.
    - distance (float): The total distance between the origin and destination locations in kilometers or miles. 
    - weight (float): The total weight of the shipment in kilograms or pounds. Make sure that it is aligned with the dimension that is used in the freight matrix.
    - volume (float): The total volume of the shipment in cubic meters or cubic feet. Make sure that it is aligned with the dimension that is used in the freight matrix.
    - pallets (int): The number of pallets used for the shipment. Make sure that it is aligned with the dimension that is used in the freight matrix.
    - loadingMeters (float): The total space in meters required by the shipment on the transport vehicle. Make sure that it is aligned with the dimension that is used in the freight matrix.

    matrix_id (str): The Freight Matrix ID from a matrix that was created on the Log-hub Platform.

    api_key (str): The Log-hub API key for accessing the freight matrix service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).
    
    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    pd.DataFrame: A pandas DataFrame containing evaluated shipments with additional
                  details such as fromLatitude, fromLongitude, toLatitude, toLongitude, costs, etc.
                  Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    float_columns = ['distance', 'weight', 'volume', 'pallets', 'loadingMeters']
    shipments_df = convert_to_float(shipments_df, float_columns, 'optional')
    mandatory_columns = {'shipmentId': 'str', 'fromLocationId': 'str', 'fromCountry': 'str', 'toLocationId': 'str', 'toCountry': 'str'}
    optional_columns = {'fromState': 'str', 'fromCity': 'str', 'fromPostalCode': 'str', 'fromStreet': 'str', 'fromZone': 'str', 'toState': 'str', 'toCity': 'str', 'toPostalCode': 'str', 'toStreet': 'str', 'toZone': 'str'}
    shipments_df = validate_and_convert_data_types(shipments_df, mandatory_columns, 'mandatory', 'shipments')
    if not shipments_df is None:
        shipments_df = validate_and_convert_data_types(shipments_df, optional_columns, 'optional', 'shipments')
    # Convert DataFrame to list of dicts for the payload, excluding NaN values in specified columns
    shipments_list = convert_df_to_dict_excluding_nan(shipments_df, float_columns)
    
    url = create_url("freightmatrixplus")
    
    headers = create_headers(api_key)
    payload = {
        "shipments": shipments_list,
        "matrix": {"matrixId": matrix_id}
    }
    payload = save_scenario_check(save_scenario, payload)

    response_data = post_method(url, payload, headers, "freight matrix plus")
    if response_data is None:
        return None
    else:
        evaluated_shipments = pd.DataFrame(response_data['evaluatedShipments'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return evaluated_shipments

def forward_freight_matrix_plus_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'freightMatrixAddresses.xlsx')
    shipments_df = pd.read_excel(data_path, sheet_name='shipments', usecols='A:U', dtype={'shipmentId': str, 'shipmentDate': str, 'fromLocationId': str, 'toLocationId': str, 'fromPostalCode': str, 'toPostalCode': str, 'distance': str, 'weight': float, 'volume': float, 'pallets': float, 'loadingMeters': float})

    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'shipments': shipments_df, 'saveScenarioParameters': save_scenario}


def reverse_freight_matrix_plus(shipments_df: pd.DataFrame, matrix_id: str, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[pd.DataFrame]:
    """
    Calculate the reverse freight matrix for a list of shipments provided in a pandas DataFrame.
    
    Parameters:
    shipments_df (pd.DataFrame): A pandas DataFrame containing the shipment details for reverse logistics. Each row represents a unique shipment, and the columns include:
        - shipmentId (str): A unique identifier for the shipment.
        - shipmentDate (str): The scheduled date and time of the shipment in UTC, in ISO 8601 format (YYYY-MM-DDTHH:MM:SSZ).
        - fromLocationId (str): Identifier for the origin location.
        - fromLatitude (float): Latitude of the origin location.
        - fromLongitude (float): Longitude of the origin location.
        - fromZone (str): A logistics zone or area at the origin, impacting shipping strategies and costs.
        - toLocationId (str): Identifier for the destination location.
        - toLatitude (float): Latitude of the destination location.
        - toLongitude (float): Longitude of the destination location.
        - toZone (str): A logistics zone or area at the destination, impacting delivery strategies and costs.
        - distance (float): The total distance between the origin and destination locations in kilometers or miles.
        - weight (float): The total weight of the shipment in kilograms or pounds, critical for cost calculations.
        - volume (float): The total volume of the shipment in cubic meters or cubic feet, affecting space utilization.
        - pallets (int): The number of pallets used for the shipment, influencing handling requirements.
        - loadingMeters (float): The total space required by the shipment on the transport vehicle, impacting vehicle utilization.

    matrix_id (str): The Freight Matrix ID from a matrix that was created on the Log-hub Platform.

    api_key (str): The Log-hub API key for accessing the reverse freight matrix service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    pd.DataFrame: A pandas DataFrame containing evaluated shipments with additional
                  details such as costs, weight class, distance class, and price per unit. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map" , "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    float_columns = ['distance', 'weight', 'volume', 'pallets', 'loadingMeters']
    shipments_df = convert_to_float(shipments_df, float_columns, 'optional')
    mandatory_columns = {'shipmentId': 'str', 'fromLocationId': 'str', 'fromLatitude': 'float', 'fromLongitude': 'float', 'toLocationId': 'str', 'toLatitude': 'float', 'toLongitude': 'float'}
    optional_columns = {'fromState': 'str', 'fromZone': 'str', 'toZone': 'str'}
    shipments_df = validate_and_convert_data_types(shipments_df, mandatory_columns, 'mandatory', 'shipments')
    if not shipments_df is None:
        shipments_df = validate_and_convert_data_types(shipments_df, optional_columns, 'optional', 'shipments')

    # Convert DataFrame to list of dicts for the payload, excluding NaN values in specified columns
    shipments_list = convert_df_to_dict_excluding_nan(shipments_df, float_columns)
    
    url = create_url("reversefreightmatrixplus")
    
    headers = create_headers(api_key)
    payload = {
        "shipments": shipments_list,
        "matrix": {"matrixId": matrix_id}
    }
    payload = save_scenario_check(save_scenario, payload)

    response_data = post_method(url, payload, headers, "reverse freight matrix plus")
    if response_data is None:
        return None
    else:
        evaluated_shipments = pd.DataFrame(response_data['evaluatedShipments'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return evaluated_shipments

def reverse_freight_matrix_plus_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'freightMatrixReverse.xlsx')
    shipments_df = pd.read_excel(data_path, sheet_name='shipments', usecols='A:O', dtype={'shipmentId': str, 'shipmentDate': str, 'fromLocationId': str, 'toLocationId': str, 'weight': float, 'volume': float, 'pallets': float, 'loadingMeters': float})

    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'shipments': shipments_df, 'saveScenarioParameters': save_scenario}


import os
import pandas as pd
import logging
import warnings
from typing import Optional, Dict, Tuple
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import convert_dates, validate_and_convert_data_types, convert_to_float, validate_boolean, convert_df_to_dict_excluding_nan
from sending_requests import post_method, create_headers, create_url, get_workspace_entities, get_method

def forward_shipment_analyzer(shipments: pd.DataFrame, cost_adjustment: pd.DataFrame, consolidation: pd.DataFrame, surcharges: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame]]:
    """
    Perform shipment analysis based on shipments, cost adjustments, consolidation settings, surcharges, and parameters.

    Parameters:
    shipments (pd.DataFrame): DataFrame containing shipment information.
        Columns and their types:
        - shipmentId (str): Unique identifier for the shipment.
        - shipmentLeg (str): The leg of the shipment.
        - fromId (str), toId (str): Identifiers for the origin and destination.
        - fromCountry (str), toCountry (str): Country codes for origin and destination.
        - fromState (str), toState (str): State information for origin and destination.
        - fromCity (str), toCity (str): City names for origin and destination.
        - fromPostalCode (str), toPostalCode (str): Postal codes for origin and destination.
        - fromStreet (str), toStreet (str): Street addresses for origin and destination.
        - fromUnLocode (str), toUnLocode (str): UN/LOCODE for origin and destination.
        - fromIataCode (str), toIataCode (str): IATA codes for origin and destination.
        - shippingDate (str), expectedDeliveryDate (str), actualDeliveryDate (str): Dates related to the shipment (Format: YYYY-MM-DD).
        - shippingMode (str): Mode of transportation (e.g., Road, Air).
        - carrier (str): Carrier information.
        - truckShipPlaneType (str): Type of transportation vehicle.
        - speedProfile (str): Speed profile used.
        - distance (float): Distance between sender and recipient. It will be calculated if not provided.
        - weight (float), volume (float), pallets (float): Shipment metrics.
        - shipmentValue (float): Value of the shipment.
        - freightCosts (float): Cost of freight.
        - benchmarkTariff (str): Benchmark tariff information.
        - surcharges (str): Any additional surcharges.

    costAdjustment (pd.DataFrame): DataFrame containing cost adjustment settings.
        Columns and their types:
        - fromIso2Country (str), toIso2Country (str): ISO country codes for adjustment applicability.
        - truckShipPlaneType (str): Type of transportation vehicle for which adjustment is applicable.
        - carrier (str): Carrier information for which adjustment is applicable.
        - benchmarkTariff (str): Benchmark tariff information.
        - factor (float): Adjustment factor.
        - flatOnTop (float): Flat fee on top of the cost.

    consolidation (pd.DataFrame): DataFrame containing consolidation settings.
        Columns and their types:
        - truckShipPlaneType (str): Type of transportation vehicle for consolidation.
        - fromIso2Country (str), toIso2Country (str): ISO country codes for consolidation applicability.
        - carrier (str): Carrier information for consolidation.
        - capacityWeight (float), capacityVolume (float), capacityPallets (float): Capacity metrics for consolidation.
        - consolidationFrequency (str): Frequency of consolidation (e.g., weekly, bi-weekly).

    surcharges (pd.DataFrame): DataFrame containing surcharges information.
        Columns and their types:
        - surcharge (str): Type of surcharge.
        - flatOnTop (float): Flat surcharge amount.

    parameters (Dict): Dictionary containing parameters for the analysis.
        Keys and their types:
        - consolidation (bool): Boolean flag to indicate if consolidation should be considered in the analysis.
        - distanceUnit (str): Distance unit "km" or "mi".

    api_key (str): Log-hub API key for accessing the shipment analyzer service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).
    
    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame]: Two pandas DataFrames containing shipment analysis and transport analysis.
                                    Returns None if the process fails.
    """

    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    shipments_mandatory_columns = {'shipmentId': 'str', 'fromId': 'str', 'toId': 'str', 'shippingDate': 'str', 'weight': 'float'}
    shipments_optional_columns = {'shipmentLeg': 'str', 'fromCountry': 'str', 'toCountry': 'str', 'fromState': 'str', 'toState': 'str',
                    'fromCity': 'str', 'toCity': 'str', 'fromPostalCode': 'str', 'toPostalCode': 'str', 'fromStreet': 'str', 'toStreet': 'str',
                    'fromUnLocode': 'str', 'toUnLocode': 'str', 'fromIataCode': 'str', 'toIataCode': 'str', 'shippingMode': 'str', 'carrier': 'str',
                    'truckShipPlaneType': 'str', 'speedProfile': 'str', 'benchmarkTariff': 'str', 'surcharges': 'str', 'expectedDeliveryDate': 'str', 'actualDeliveryDate': 'str'}
    shipments_optional_floats = ['distance', 'volume', 'pallets', 'shipmentValue', 'freightCosts']

    shipments = validate_and_convert_data_types(shipments, shipments_mandatory_columns, 'mandatory', 'shipments')
    if not shipments is None:
        shipments = validate_and_convert_data_types(shipments, shipments_optional_columns, 'optional', 'shipments')
        if not shipments is None:
            shipments = convert_dates(shipments, ['shippingDate', 'expectedDeliveryDate', 'actualDeliveryDate'])
            shipments = convert_to_float(shipments, shipments_optional_floats, 'optional')
            shipments = convert_df_to_dict_excluding_nan(shipments, shipments_optional_floats)
    cost_adjustment_optional_columns = {'fromIso2Country': 'str', 'toIso2Country': 'str', 'truckShipPlaneType': 'str', 'carrier':'str', 'benchmarkTariff': 'str'}
    cost_adjustment = validate_and_convert_data_types(cost_adjustment, cost_adjustment_optional_columns, 'optional', 'cots adjustment')
    if not cost_adjustment is None:
        cost_adjustment = convert_to_float(cost_adjustment, ['factor', 'flatOnTop'], 'optional')
        cost_adjustment = convert_df_to_dict_excluding_nan(cost_adjustment, ['factor', 'flatOnTop'])
    consolidation_optional_columns = {'fromIso2Country': 'str', 'toIso2Country': 'str', 'truckShipPlaneType': 'str', 'carrier':'str', 'consolidationFrequency': 'str'}
    consolidation = validate_and_convert_data_types(consolidation, consolidation_optional_columns, 'optional', 'consolidation')
    if not consolidation is None:
        consolidation = convert_to_float(consolidation, ['capacityWeight', 'capacityVolume', 'capacityPallets'], 'optional')
        consolidation = convert_df_to_dict_excluding_nan(consolidation, ['capacityWeight', 'capacityVolume', 'capacityPallets'])
    surcharges_optional_columns = {'surcharge': 'str'}
    surcharges = validate_and_convert_data_types(surcharges, surcharges_optional_columns, 'optional', 'surcharges')
    if not surcharges is None:
        surcharges = convert_to_float(surcharges, ['flatOnTop'], 'optional')
        surcharges = convert_df_to_dict_excluding_nan(surcharges, ['flatOnTop'])

    # Validate boolean parameter
    if 'consolidation' in parameters and not validate_boolean(parameters['consolidation']):
        logging.error("Invalid type for 'consolidation' in parameters. It should be boolean.")
        return None
    if any(df is None for df in [shipments, cost_adjustment, consolidation,surcharges]):
        return None

    url = create_url("shipmentanalyzerpluslongrun")
    
    headers = create_headers(api_key)

    payload = {
        "shipments": shipments,
        "costAdjustment": cost_adjustment,
        "consolidation": consolidation,
        "surcharges": surcharges,
        "parameters": parameters
    }
    payload = save_scenario_check(save_scenario, payload)
    response_data = post_method(url, payload, headers, "shipment analyzer plus")
    if response_data is None:
        return None
    else:
        result = response_data['result']
        get_method_result = get_method(result['apiServer'], result['url'], {"authorization": f"apikey {api_key}"}, "shipment analyzer plus")
        if get_method_result is None:
            return None
        else:
            shipments_df = pd.DataFrame(get_method_result['shipments'])
            transports_df = pd.DataFrame(get_method_result['transports'])
            if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
                create_buttons()
            if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
                logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
            return shipments_df, transports_df
           
def forward_shipment_analyzer_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'shipmentAnalyzerAddresses.xlsx')
    shipments_df = pd.read_excel(data_path, sheet_name='shipments', usecols='A:AH').fillna("")
    transport_costs_adjustments_df = pd.read_excel(data_path, sheet_name='transportCostAdjustments', usecols='A:H').fillna("")
    consolidation_df = pd.read_excel(data_path, sheet_name='consolidation', usecols='A:I').fillna("")
    surcharges_df = pd.read_excel(data_path, sheet_name='surcharges', usecols='A:C').fillna("")

    parameters = {
        "consolidation": False,
        "distanceUnit": "km"
    }

    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }

    return {'shipments': shipments_df, 'transportCostAdjustments': transport_costs_adjustments_df, 'consolidation': consolidation_df, 'surcharges': surcharges_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


def reverse_shipment_analyzer(shipments: pd.DataFrame, cost_adjustment: pd.DataFrame, consolidation: pd.DataFrame, surcharges: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame]]:
    """
    Perform reverse shipment analysis based on shipments, cost adjustments, consolidation settings, surcharges, and parameters.

    Parameters:
    shipments (pd.DataFrame): DataFrame containing reverse shipment information.
        Columns and their types:
        - shipmentId (str): Unique identifier for the shipment.
        - shipmentLeg (str): The leg of the shipment.
        - fromId (str): Identifier for the origin.
        - fromLatitude (float): Latitude for the origin.
        - fromLongitude (float): Longitude for the origin.
        - toId (str): Identifier for the destination.
        - toLatitude (float): Latitude for the destination.
        - toLongitude (float): Longitude for the destination.
        - shippingDate (str): Date of shipping (Format: YYYY-MM-DD).
        - expectedDeliveryDate (str): Expected delivery date (Format: YYYY-MM-DD).
        - actualDeliveryDate (str): Actual delivery date (Format: YYYY-MM-DD).
        - shippingMode (str): Mode of transportation (e.g., Road, Air).
        - carrier (str): Carrier information.
        - truckShipPlaneType (str): Type of transportation vehicle.
        - speedProfile (str): Speed profile used.
        - distance (float): Distance between sender and recipient. It will be calculated if not provided.
        - weight (float): Weight of the shipment.
        - volume (float): Volume of the shipment.
        - pallets (float): Number of pallets.
        - shipmentValue (float): Value of the shipment.
        - freightCosts (float): Cost of freight.
        - benchmarkTariff (str): Benchmark tariff information.
        - surcharges (str): Any additional surcharges.

    costAdjustment (pd.DataFrame): DataFrame containing cost adjustment settings.
        Columns and their types:
        - fromIso2Country (str), toIso2Country (str): ISO country codes for adjustment applicability.
        - truckShipPlaneType (str): Type of transportation vehicle for which adjustment is applicable.
        - carrier (str): Carrier information for which adjustment is applicable.
        - benchmarkTariff (str): Benchmark tariff information.
        - factor (float): Adjustment factor.
        - flatOnTop (float): Flat fee on top of the cost.

    consolidation (pd.DataFrame): DataFrame containing consolidation settings.
        Columns and their types:
        - truckShipPlaneType (str): Type of transportation vehicle for consolidation.
        - fromIso2Country (str), toIso2Country (str): ISO country codes for consolidation applicability.
        - carrier (str): Carrier information for consolidation.
        - capacityWeight (float), capacityVolume (float), capacityPallets (float): Capacity metrics for consolidation.
        - consolidationFrequency (str): Frequency of consolidation (e.g., weekly, bi-weekly).

    surcharges (pd.DataFrame): DataFrame containing surcharges information.
        Columns and their types:
        - surcharge (str): Type of surcharge.
        - flatOnTop (float): Flat surcharge amount.

    parameters (Dict): Dictionary containing parameters for the analysis.
        Keys and their types:
        - consolidation (bool): Boolean flag to indicate if consolidation should be considered in the analysis.
        - distanceUnit (str): Distance unit "km" or "mi".

    api_key (str): API key for accessing the reverse shipment analyzer service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).
    
    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.
    Returns:
    Tuple[pd.DataFrame, pd.DataFrame]: Two pandas DataFrames containing reverse shipment analysis and transport analysis.
                                      The first DataFrame ('shipments') includes details such as shipment IDs, locations, dates, shipment metrics, and calculated emissions and costs.
                                      The second DataFrame ('transports') includes aggregated transport metrics such as total weights, volumes, costs, and emissions for each transport ID.
                                      Returns None if the process fails.
    """

    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    shipments_mandatory_columns = {'shipmentId': 'str', 'fromId': 'str', 'toId': 'str', 'shippingDate': 'str', 'weight': 'float', 'fromLatitude': 'float', 'fromLongitude': 'float', 'toLatitude': 'float', 'toLongitude': 'float'}
    shipments_optional_columns = {'shipmentLeg': 'str', 'fromCountry': 'str', 'toCountry': 'str', 'fromState': 'str', 'toState': 'str',
                    'fromCity': 'str', 'toCity': 'str', 'fromPostalCode': 'str', 'toPostalCode': 'str', 'fromStreet': 'str', 'toStreet': 'str',
                    'fromUnLocode': 'str', 'toUnLocode': 'str', 'fromIataCode': 'str', 'toIataCode': 'str', 'shippingMode': 'str', 'carrier': 'str',
                    'truckShipPlaneType': 'str', 'speedProfile': 'str', 'benchmarkTariff': 'str', 'surcharges': 'str', 'expectedDeliveryDate': 'str', 'actualDeliveryDate': 'str'}
    shipments_optional_floats = ['volume', 'pallets', 'shipmentValue', 'freightCosts', 'distance']

    shipments = validate_and_convert_data_types(shipments, shipments_mandatory_columns, 'mandatory', 'shipments')
    if not shipments is None:
        shipments = validate_and_convert_data_types(shipments, shipments_optional_columns, 'optional', 'shipments')
        if not shipments is None:
            shipments = convert_dates(shipments, ['shippingDate', 'expectedDeliveryDate', 'actualDeliveryDate'])
            shipments = convert_to_float(shipments, shipments_optional_floats, 'optional')
            shipments = convert_df_to_dict_excluding_nan(shipments, shipments_optional_floats)
    cost_adjustment_optional_columns = {'fromIso2Country': 'str', 'toIso2Country': 'str', 'truckShipPlaneType': 'str', 'carrier':'str', 'benchmarkTariff': 'str'}
    cost_adjustment = validate_and_convert_data_types(cost_adjustment, cost_adjustment_optional_columns, 'optional', 'cots adjustment')
    if not cost_adjustment is None:
        cost_adjustment = convert_to_float(cost_adjustment, ['factor', 'flatOnTop'], 'optional')
        cost_adjustment = convert_df_to_dict_excluding_nan(cost_adjustment, ['factor', 'flatOnTop'])
    consolidation_optional_columns = {'fromIso2Country': 'str', 'toIso2Country': 'str', 'truckShipPlaneType': 'str', 'carrier':'str', 'consolidationFrequency': 'str'}
    consolidation = validate_and_convert_data_types(consolidation, consolidation_optional_columns, 'optional', 'consolidation')
    if not consolidation is None:
        consolidation = convert_to_float(consolidation, ['capacityWeight', 'capacityVolume', 'capacityPallets'], 'optional')
        consolidation = convert_df_to_dict_excluding_nan(consolidation, ['capacityWeight', 'capacityVolume', 'capacityPallets'])
    surcharges_optional_columns = {'surcharge': 'str'}
    surcharges = validate_and_convert_data_types(surcharges, surcharges_optional_columns, 'optional', 'surcharges')
    if not surcharges is None:
        surcharges = convert_to_float(surcharges, ['flatOnTop'], 'optional')
        surcharges = convert_df_to_dict_excluding_nan(surcharges, ['flatOnTop'])

    # Validate boolean parameter
    if 'consolidation' in parameters and not validate_boolean(parameters['consolidation']):
        logging.error("Invalid type for 'consolidation' in parameters. It should be boolean.")
        return None
    if any(df is None for df in [shipments, cost_adjustment, consolidation,surcharges]):
        return None

    url = create_url("reverseshipmentanalyzerpluslongrun")
    
    headers = create_headers(api_key)

    payload = {
        "shipments": shipments,
        "costAdjustment": cost_adjustment,
        "consolidation": consolidation,
        "surcharges": surcharges,
        "parameters": parameters
    }
    payload = save_scenario_check(save_scenario, payload)
    
    response_data = post_method(url, payload, headers, "reverse shipment analyzer plus")
    if response_data is None:
        return None
    else:
        result = response_data['result']
        get_method_result = get_method(result['apiServer'], result['url'], {"authorization": f"apikey {api_key}"}, "shipment analyzer plus")
        if get_method_result is None:
            return None
        else:
            shipments_df = pd.DataFrame(get_method_result['shipments'])
            transports_df = pd.DataFrame(get_method_result['transports'])
            if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
                create_buttons()
            if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
                logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
            return shipments_df, transports_df
            
def reverse_shipment_analyzer_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'shipmentAnalyzerReverse.xlsx')
    shipments_df = pd.read_excel(data_path, sheet_name='shipments', usecols='A:X').fillna("")
    transport_costs_adjustments_df = pd.read_excel(data_path, sheet_name='transportCostAdjustments', usecols='A:H').fillna("")
    consolidation_df = pd.read_excel(data_path, sheet_name='consolidation', usecols='A:I').fillna("")
    surcharges_df = pd.read_excel(data_path, sheet_name='surcharges', usecols='A:C').fillna("")

    parameters = {
        "consolidation": False,
        "distanceUnit": "km"
    }

    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'shipments': shipments_df, 'transportCostAdjustments': transport_costs_adjustments_df, 'consolidation': consolidation_df, 'surcharges': surcharges_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


import os
import pandas as pd
import warnings
from typing import Optional, Dict, Tuple
import logging
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import validate_and_convert_data_types
from sending_requests import post_method, create_headers, create_url, get_workspace_entities

def forward_fixed_center_of_gravity(customers: pd.DataFrame, fixed_centers: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame]]:
    """
    Calculate fixed center of gravity based on a list of customers and their weights, and predefined fixed centers.

    This function takes a DataFrame of customers with weights, a DataFrame of fixed centers (if any), and a set of parameters,
    along with an API key, and performs fixed center of gravity calculation using the Log-hub service.

    Parameters:
    customers (pd.DataFrame): A pandas DataFrame containing customers with their weights.
        Each row should contain:
        - id (number): Identifier.
        - name (str): Name. Must have at least 1 character.
        - country (str): Country code. Must have at least 1 character.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        - weight (float): Weight.

    fixed_centers (pd.DataFrame): A pandas DataFrame containing predefined fixed centers.
        Each row should contain:
        - id (number): Identifier.
        - name (str): Name. Must have at least 1 character.
        - country (str): Country code. Must have at least 1 character.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        This DataFrame can be empty.

    parameters (Dict): A dictionary containing parameters like numberOfCenters (number) and distanceUnit (enum: "km" or "mi").

    api_key (str): The Log-hub API key for accessing the fixed center of gravity service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame]: A tuple of two pandas DataFrames. The first DataFrame contains the 
                                       assigned geocodes with their respective centers, and the second 
                                       DataFrame contains the details of the centers.
                                       Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])
    
    customer_mandatory_columns = {'name': 'str', 'country': 'str',  'weight': 'float'}
    customer_optional_columns = {'id': 'float', 'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str',}
    fixed_center_mandatory_columns = {'name': 'str', 'country': 'str'}
    fixed_center_optional_columns = {'id': 'float', 'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str'}

    # Validate and convert data types for customers and fixed centers
    customers = validate_and_convert_data_types(customers, customer_mandatory_columns, 'mandatory', 'customers')
    if not customers is None:
        customers = validate_and_convert_data_types(customers, customer_optional_columns, 'optional', 'customers')

    fixed_centers = validate_and_convert_data_types(fixed_centers, fixed_center_mandatory_columns, 'mandatory', 'fixed centers') if not fixed_centers.empty else fixed_centers
    if not fixed_centers is None:
        fixed_centers = validate_and_convert_data_types(fixed_centers, fixed_center_optional_columns, 'optional', 'fixed centers') if not fixed_centers.empty else fixed_centers
    if customers is None or (not fixed_centers.empty and fixed_centers is None):
        return None
    
    url = create_url("fixedcenterofgravity")
    
    headers = create_headers(api_key)

    payload = {
        "customers": customers.to_dict(orient='records'),
        "fixedCenters": fixed_centers.to_dict(orient='records') if not fixed_centers.empty else [],
        "parameters": parameters
    }
    payload = save_scenario_check(save_scenario, payload)
    response_data = post_method(url, payload, headers, "fixedcenterofgravity")
    if response_data is None:
        return None
    else:
        assigned_geocodes_df = pd.DataFrame(response_data['assignedGeocodes'])
        centers_df = pd.DataFrame(response_data['centers'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return assigned_geocodes_df, centers_df
            
def forward_fixed_center_of_gravity_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'FixedCOGSampleDataAddresses.xlsx')
    customers_df = pd.read_excel(data_path, sheet_name='customers', usecols='A:H', dtype={'postalCode': str})
    fixedCenters_df = pd.read_excel(data_path, sheet_name='fixedCenters', usecols='A:G', dtype={'postalCode': str})

    parameters = {
        "numberOfCenters": 5,
        "distanceUnit": "km"
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'customers': customers_df, 'fixedCenters': fixedCenters_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


def reverse_fixed_center_of_gravity(customers: pd.DataFrame, fixed_centers: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame]]:
    """
    Calculate reverse fixed center of gravity based on a list of customers and their weights, and predefined fixed centers.

    This function takes a DataFrame of customers with weights, a DataFrame of fixed centers, and a set of parameters,
    along with an API key, and performs reverse fixed center of gravity calculation using the Log-hub service.

    Parameters:
    customers (pd.DataFrame): A pandas DataFrame containing customers with their weights.
        Each row should contain:
        - id (number): Identifier.
        - name (str): Name.
        - latitude (float): Latitude of the customer location.
        - longitude (float): Longitude of the customer location.
        - weight (float): Weight.

    fixed_centers (pd.DataFrame): A pandas DataFrame containing predefined fixed centers.
        Each row should contain:
        - id (number): Identifier.
        - name (str): Name.
        - latitude (float): Latitude of the fixed center location.
        - longitude (float): Longitude of the fixed center location.

    parameters (Dict): A dictionary containing parameters like numberOfCenters (number) and distanceUnit (enum: "km" or "mi").

    api_key (str): The Log-hub API key for accessing the reverse fixed center of gravity service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).
    
    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame]: A tuple of two pandas DataFrames. The first DataFrame contains the 
                                       assigned geocodes with their respective centers, and the second 
                                       DataFrame contains the details of the centers.
                                       Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])
    customer_mandatory_columns = {
        'name': 'str', 'latitude': 'float', 'longitude': 'float', 'weight': 'float'
    }
    customer_optional_columns = {'id': 'float'}

    fixed_center_mandatory_columns = {
        'name': 'str', 'latitude': 'float', 'longitude': 'float'
    }
    fixed_center_optional_columns = {'id': 'float'}

    # Validate and convert data types for customers and fixed centers
    customers = validate_and_convert_data_types(customers, customer_mandatory_columns, 'mandatory', 'customers')
    if not customers is None:
        customers = validate_and_convert_data_types(customers, customer_optional_columns, 'optional', 'customers')

    fixed_centers = validate_and_convert_data_types(fixed_centers, fixed_center_mandatory_columns, 'mandatory', 'fixed centers') if not fixed_centers.empty else fixed_centers
    if not fixed_centers is None:
        fixed_centers = validate_and_convert_data_types(fixed_centers, fixed_center_optional_columns, 'optional', 'fixed centers') if not fixed_centers.empty else fixed_centers
    if customers is None or (not fixed_centers.empty and fixed_centers is None):
        return None

    url = create_url("reversefixedcenterofgravity")
    
    headers = create_headers(api_key)

    payload = {
        "customers": customers.to_dict(orient='records'),
        "fixedCenters": fixed_centers.to_dict(orient='records'),
        "parameters": parameters
    }
    payload = save_scenario_check(save_scenario, payload)
    response_data = post_method(url, payload, headers, "reverse fixed center of gravity")
    if response_data is None:
        return None
    else:
        assigned_geocodes_df = pd.DataFrame(response_data['assignedGeocodes'])
        centers_df = pd.DataFrame(response_data['centers'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return assigned_geocodes_df, centers_df

def reverse_fixed_center_of_gravity_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'FixedCOGSampleDataReverse.xlsx')
    customers_df = pd.read_excel(data_path, sheet_name='customers', usecols='A:E')
    fixedCenters_df = pd.read_excel(data_path, sheet_name='fixedCenters', usecols='A:D')

    parameters = {
        "numberOfCenters": 5,
        "distanceUnit": "km"
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'customers': customers_df, 'fixedCenters': fixedCenters_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


import os
import requests
import pandas as pd
import logging
import json
import base64
from urllib.parse import urlparse
logging.basicConfig(level=logging.INFO)

# Configure logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

def read_table(table_link, email, api_key):
    """
    Reads data from a specified table on the Log-hub platform and converts it into a pandas DataFrame. The function uses basic authentication for secure access and formats the DataFrame columns based on the table's metadata.

    This function is designed to simplify the process of retrieving data from a Log-hub table. It fetches both the data and its associated metadata to ensure the data types in the pandas DataFrame match those in the Log-hub table.

    Parameters:
    table_link (str): The URL link to the specific table on the Log-hub platform. This link should direct to the API endpoint where the table data can be accessed.
    email (str): The email address used for Log-hub authentication.
    api_key (str): The API key associated with your Log-hub account for authentication.

    Returns:
    pd.DataFrame/None: A pandas DataFrame containing the table's data formatted according to its metadata. Returns None if the data retrieval is unsuccessful, or if any errors occur during the process.

    Handling Errors:
    - If there are any issues in fetching the metadata or the data (like network issues, authentication failures, or incorrect URLs), the function logs an error and returns None.
    - Errors are also logged and None is returned in case of any unexpected issues during data type conversion or DataFrame creation.

    Example Usage:
    table_link = "https://production.supply-chain-apps.log-hub.com/api/v1/datasets/.../tables/.../rows"
    email = "your_email@example.com"
    api_key = "your_api_key"
    dataframe = read_table(table_link, email, api_key)

    Note: Replace the 'table_link', 'email', and 'api_key' with actual values from your Log-hub account.
    """


    def js_to_pd_dtype(js_dtype):
        """
        Maps JavaScript data types to pandas data types.

        Parameters:
        js_dtype (str): The JavaScript data type.

        Returns:
        dtype: The corresponding pandas data type.
        """
        return {
            'string': 'object',
            'number': 'float64',
            'date': 'datetime64[ns]',
            'boolean': 'bool'
        }.get(js_dtype, 'object')  # default to object if unknown type

    # Prepare the basic auth string
    auth_str = f"{email}:{api_key}"
    encoded_auth_str = base64.b64encode(auth_str.encode()).decode()

    # Headers for the request, including the basic authentication
    headers = {
        "Authorization": f"Basic {encoded_auth_str}"
    }

    # Construct metadata link from the table link
    metadata_link = '/'.join(table_link.split('/')[:-2])

    try:
        # Get the table metadata
        metadata_response = requests.get(metadata_link, headers=headers)
        if metadata_response.status_code != 200:
            logging.error(f"HTTP Error in metadata request: {metadata_response.status_code} - {metadata_response.text}")
            return None

        metadata_json = metadata_response.json()
        if 'data' not in metadata_json or not metadata_json['data']:
            logging.error("Metadata JSON does not contain 'data' key or it's empty")
            return None

        column_types = {col['propertyName']: js_to_pd_dtype(col['dataType']) for col in metadata_json['data'][0]['columns']}

        # Get the table data
        data_response = requests.get(table_link, headers=headers)
        if data_response.status_code != 200:
            logging.error(f"HTTP Error in data request: {data_response.status_code} - {data_response.text}")
            return None

        # Create DataFrame and format columns
        df = pd.DataFrame(data_response.json())
        for col, dtype in column_types.items():
            if col in df.columns:
                df[col] = df[col].astype(dtype)

        return df

    except requests.exceptions.RequestException as e:
        logging.error(f"Request Exception: {e}")
        return None
    except KeyError as e:
        logging.error(f"Key Error: {e}")
        return None
    except Exception as e:
        logging.error(f"Unexpected error: {e}")
        return None
    

# Configure logging
logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

def extract_ids_from_link(table_link):
    """
    Extracts dataset_id and table_id from the provided table link.

    Parameters:
    table_link (str): The link to the table.

    Returns:
    tuple: The dataset_id and table_id.
    """
    parsed_url = urlparse(table_link)
    path_segments = parsed_url.path.split('/')
    # Finding indices for 'datasets' and 'tables' in the path
    datasets_index = path_segments.index('datasets')
    tables_index = path_segments.index('tables')

    # The dataset_id is the segment following 'datasets'
    dataset_id = path_segments[datasets_index + 1]
    # The table_id is the segment following 'tables'
    table_id = path_segments[tables_index + 1]
    return dataset_id, table_id


def pd_to_js_dtype(pd_dtype):
    """
    Maps pandas data types to JavaScript data types.
    """
    if pd.api.types.is_string_dtype(pd_dtype):
        return 'string'
    elif pd.api.types.is_numeric_dtype(pd_dtype):
        return 'number'
    elif pd.api.types.is_datetime64_any_dtype(pd_dtype):
        return 'date'
    elif pd.api.types.is_bool_dtype(pd_dtype):
        return 'boolean'
    else:
        return 'string'  # default type

def create_metadata(dataframe, provided_metadata=None):
    """
    Creates metadata for the columns of the dataframe.

    Parameters:
    dataframe (pd.DataFrame): The dataframe to create metadata for.
    provided_metadata (dict, optional): Provided metadata to override defaults.

    Returns:
    list: A list of metadata dictionaries for each column.
    """
    metadata = []
    for col in dataframe.columns:
        col_metadata = {
            'name': col,
            'propertyName': col,
            'dataType': pd_to_js_dtype(dataframe[col].dtype),
            'format': 'General'
        }
        # Overriding with any provided metadata for the column
        if provided_metadata and col in provided_metadata:
            col_metadata.update(provided_metadata[col])
        metadata.append(col_metadata)
    return metadata

def update_table(table_link, dataframe, metadata, email, api_key):
    """
    Updates a table in a specific dataset on the Log-hub platform using data from a pandas DataFrame.

    This function is designed to easily push updates from a local DataFrame to a table hosted on the Log-hub platform. It requires a link to the specific table you wish to update, the DataFrame containing your data, metadata to describe the data structure, and your Log-hub authentication details.

    Parameters:
    table_link (str): A URL pointing to the specific table you want to update on the Log-hub platform. 
                      This link should be in the format provided by Log-hub for accessing tables.
    dataframe (pd.DataFrame): The DataFrame containing the data you want to upload. Ensure that the 
                              DataFrame structure (columns and data types) matches the structure of 
                              the table you're updating.
    metadata (dict): A dictionary that describes the structure and properties of your table and its columns. 
                     It can include the name of the table and detailed descriptions for each column.
                     If column metadata is not provided, it will be generated based on the DataFrame's structure.
    email (str): Your email address used for Log-hub authentication.
    api_key (str): Your Log-hub API key.

    The 'metadata' Dictionary:
    - 'table_name' (optional): A string representing the name of the table. Defaults to "Table 01" if not provided.
    - 'columns' (optional): A list of dictionaries, each representing a column in your table. Each dictionary should have:
        * 'name': The name of the column.
        * 'propertyName': A property name for the column, typically the same as 'name'.
        * 'dataType': The type of data in the column ('string', 'number', 'date', 'boolean').
        * 'format': The format of the data (e.g., 'General', 'Date').

    Example of 'metadata' dictionary:
    {
        'table_name': 'SalesData',  # Optional
        'columns': [  # Optional, but recommended for clarity
            {
                'name': 'ProductID',
                'propertyName': 'ProductID',
                'dataType': 'string',
                'format': 'General'
            },
            {
                'name': 'SaleAmount',
                'propertyName': 'SaleAmount',
                'dataType': 'number',
                'format': 'Currency'
            }
            # Add more columns as needed
        ]
    }

    Example Usage:
    table_link = "https://production.supply-chain-apps.log-hub.com/api/v1/datasets/.../tables/.../rows"
    dataframe = pd.DataFrame({'ProductID': ['P123', 'P456'], 'SaleAmount': [150, 200]})
    metadata = {
        'table_name': 'SalesData',
        'columns': [
            {'name': 'ProductID', 'propertyName': 'ProductID', 'dataType': 'string', 'format': 'General'},
            {'name': 'SaleAmount', 'propertyName': 'SaleAmount', 'dataType': 'number', 'format': 'Currency'}
        ]
    }
    update_table(table_link, dataframe, metadata, 'your_email@example.com', 'your_api_key')
    
    Note: The function handles the connection and data transfer to Log-hub. It logs success or error messages and returns None.
    """
    dataset_id, table_id = extract_ids_from_link(table_link)

    # Prepare the basic auth string
    auth_str = f"{email}:{api_key}"
    encoded_auth_str = base64.b64encode(auth_str.encode()).decode()

    # Headers for the request
    headers = {
        "Authorization": f"Basic {encoded_auth_str}",
        "Content-Type": "application/json"
    }

    # Set table name in metadata
    table_name = metadata.get('table_name', "Table 01")

    # Create metadata if not provided or if it does not contain column information
    if not metadata or 'columns' not in metadata:
        metadata = {'columns': create_metadata(dataframe, provided_metadata=metadata)}

    # Prepare data for update including the table name and columns metadata
    update_data = {
        "name": table_name,
        "columns": metadata['columns'],
        "rows": dataframe.to_dict(orient='records')
    }

    DEFAULT_LOG_HUB_API_SERVER = "https://production.supply-chain-apps.log-hub.com"
    LOG_HUB_API_SERVER = os.getenv('LOG_HUB_API_SERVER', DEFAULT_LOG_HUB_API_SERVER)
    url = f"{LOG_HUB_API_SERVER}/api/v1/datasets/{dataset_id}/tables/{table_id}"

    try:
        # Send PATCH request
        response = requests.patch(url, headers=headers, data=json.dumps(update_data))

        # Check if the request was successful
        if response.status_code == 200:
            logging.info("Table updated successfully.")
            # return response.json()
            return None
        else:
            logging.error(f"Failed to update table. Status code: {response.status_code} - {response.text}")
            return None

    except requests.exceptions.RequestException as e:
        logging.error(f"Request Exception: {e}")
        return None



import os
import pandas as pd
import logging
from typing import Optional, Dict
import warnings
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import validate_and_convert_data_types
from sending_requests import post_method, create_headers, create_url, get_workspace_entities

logging.basicConfig(level=logging.INFO)

def forward_distance_calculation(address_pairs: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[pd.DataFrame]:
    """
    Calculate distances and durations between pairs of addresses.

    This function takes a DataFrame of address pairs (sender and recipient) and a set of parameters,
    along with an API key, and performs distance calculation using the Log-hub distance calculation service. 

    Parameters:
    address_pairs (pd.DataFrame): A pandas DataFrame containing pairs of addresses.
        Each row should contain:
        - senderCountry (str): Sender country code. Must have at least 1 character.
        - senderState (str): Sender state code.
        - senderPostalCode (str): Sender postal code.
        - senderCity (str): Sender city name.
        - senderStreet (str): Sender street name with house number.
        - recipientCountry (str): Recipient country code. Must have at least 1 character.
        - recipientState (str): Recipient state code.
        - recipientPostalCode (str): Recipient postal code.
        - recipientCity (str): Recipient city name.
        - recipientStreet (str): Recipient street name with house number.

    parameters (Dict): A dictionary containing parameters like distanceUnit (enum: "km" or "mi"),
                        durationUnit (enum: "min" or "sec" or "h"), vehicleType (enum: "truck" or "car") and 
                        routePreference (enum: 'recommended' or 'shortest' or 'fastest').

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    api_key (str): The Log-hub API key for accessing the distance calculation service.

    Returns:
    pd.DataFrame: A pandas DataFrame containing the results of the distance calculations. 
                  Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])
    mandatory_columns = {'senderCountry': 'str', 'recipientCountry': 'str'}
    optional_columns = {'senderState': 'str', 'senderPostalCode': 'str', 'senderCity': 'str', 'senderStreet': 'str', 'recipientState': 'str',
                        'recipientPostalCode': 'str', 'recipientCity': 'str', 'recipientStreet': 'str'}

    address_pairs = validate_and_convert_data_types(address_pairs, mandatory_columns, 'mandatory', 'addresses')
    if not address_pairs is None:
        address_pairs = validate_and_convert_data_types(address_pairs, optional_columns, 'optional', 'addresses')
    if address_pairs is None:
        return None

    url = create_url("distancecalculation")
    
    headers = create_headers(api_key)

    payload = {
        "addresses": address_pairs.to_dict(orient='records'), 
        "parameters": parameters,
        }
    
    payload = save_scenario_check(save_scenario, payload)

    response_data = post_method(url, payload, headers, "distance calculation")
    if response_data is None:
        return None
    else:
        distances_df = pd.DataFrame(response_data)
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return distances_df
    
def forward_distance_calculation_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'DistanceCalcSampleDataAddresses.xlsx')
    addresses_df = pd.read_excel(data_path, sheet_name='addresses', usecols='A:J')

    parameters = {
        "distanceUnit": "km",
        "durationUnit": "min",
        "vehicleType": "car",
        "routePreference": "recommended"
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'address_data': addresses_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}



def reverse_distance_calculation(geocodes: pd.DataFrame, parameters: Dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[pd.DataFrame]:
    """
    Calculate distances and durations between pairs of locations based on coordinates.

    This function takes a DataFrame of geocode pairs (sender and recipient) with their locations and coordinates,
    a set of parameters, along with an API key, and performs reverse distance calculation using the 
    Log-hub reverse distance calculation service.

    Parameters:
    geocodes (pd.DataFrame): A pandas DataFrame containing pairs of geocodes with sender and recipient information.
        Each row should contain:
        - senderLocation (str): Sender location. Must have at least 1 character.
        - senderLatitude (float): Sender latitude.
        - senderLongitude (float): Sender longitude.
        - recipientLocation (str): Recipient location. Must have at least 1 character.
        - recipientLatitude (float): Recipient latitude.
        - recipientLongitude (float): Recipient longitude.

    parameters (Dict): A dictionary containing parameters like distanceUnit (enum: "km" or "mi"),
                        durationUnit (enum: "min" or "sec" or "h"), vehicleType (enum: "truck" or "car") and 
                        routePreference (enum: 'recommended' or 'shortest' or 'fastest').

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                        'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    api_key (str): The Log-hub API key for accessing the reverse distance calculation service.

    Returns:
    pd.DataFrame: A pandas DataFrame containing the results of the reverse distance calculations.
                  Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    mandatory_columns = {
            'senderLocation': 'str', 'senderLatitude': 'float', 'senderLongitude': 'float',
            'recipientLocation': 'str', 'recipientLatitude': 'float', 'recipientLongitude': 'float'
        }
    
    # Validate and convert data types
    geocodes = validate_and_convert_data_types(geocodes, mandatory_columns, 'mandatory', 'coordinates')
    if geocodes is None:
        return None

    url = create_url("reversedistancecalculation")
    
    headers = create_headers(api_key)
    payload = {
        "geocodes": geocodes.to_dict(orient='records'),
        "parameters": parameters
    }
    payload = save_scenario_check(save_scenario, payload)

    response_data = post_method(url, payload, headers, "reverse distance calculation")
    if response_data is None:
        return None
    else:
        distances_df = pd.DataFrame(response_data)
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return distances_df

def reverse_distance_calculation_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'DistanceCalcSampleDataReverse.xlsx')
    geocode_data_df = pd.read_excel(data_path, sheet_name='coordinates', usecols='A:F')

    parameters = {
        "distanceUnit": "km",
        "durationUnit": "min",
        "vehicleType": "car",
        "routePreference": "recommended"
    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'geocode_data': geocode_data_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


import os
import pandas as pd
import logging
import warnings
logging.basicConfig(level=logging.INFO)
from typing import Optional, Tuple
import sys
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import exclude_nan_depending_on_dtype, convert_to_float, validate_and_convert_data_types, convert_df_to_dict_excluding_nan
from sending_requests import post_method, create_headers, create_url, get_method, get_workspace_entities

def forward_network_design_plus(factories: pd.DataFrame, warehouses: pd.DataFrame, customers: pd.DataFrame, product_segments: pd.DataFrame, transport_costs: pd.DataFrame, transport_costs_rules: pd.DataFrame, stepwise_function_weight: pd.DataFrame, stepwise_function_volume: pd.DataFrame, distance_limits: pd.DataFrame, parameters: dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
    """
    Perform network design plus based on factories, warehouses, customers, product segments, transport costs, transport costs rules, stepwise functions for weight and volume, and distance limits.

    This function takes DataFrames of factories, warehouses, customers, product segments, transport costs, transport costs rules, stepwise functions for weight and volume, distance limits and an API key, and performs forward network design plus using the Log-hub service. 

    Parameters:
    factories (pd.DataFrame): A pandas DataFrame containing the factories information for the optimization run.
        Required columns and their types:
        - name (str): Factory name.
        - country (str): Country code or country name.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        - outboundFtlCapacityWeight (number): Maximum truck capacity regarding the weight dimension for outbound transports.
        - outboundFtlCapacityVolume (number): Maximum truck capacity regarding the volume dimension for outbound transports.
        - minimumFtlCosts (number): Minimum costs of a full truck.
        - outboundFtlCost1DistanceUnit (number): Costs per km/mi for the first distance unit.
        - outboundFtlCost1000DistanceUnit (number): Costs per km/mi for the 1000. distance unit.
        - minimumLtlCosts (number): Minimum costs of an LTL truck.
        - outboundCostsHalfFtlPercent (number): Costs for 1/2 truck.
    warehouses (pd.DataFrame): A pandas DataFrame containing the warehouses information for the optimization run.
        - name (str): Warehouse name.
        - country (str): Country code or country name.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        - fixed (0 or 1): If the warehouse should be in every case part of the result, then mark it as fixed (1).
        - minWeight (number): The weight that should be at least assigned to this warehouse.
        - maxWeight (number): The maximum weight capacity of the warehouse.
        - penaltyCostsWeight (number): Penalty costs per weight unit. 
        - minVolume (number): The volume that should be at least assigned to this warehouse.
        - maxVolume (number): The maximum volume capacity of the warehouse.
        - penaltyCostsVolume (number): Penalty costs per volume unit. 
        - fixedCosts (number):  Fixed warehouse costs.
        - costsPerWeightUnit (number): Costs by weight unit.
        - costsPerVolumeUnit (number): Costs by volume unit.
        - outboundFtlCapacityWeight (number): Maximum truck capacity regarding the weight dimension for outbound transports.
        - outboundFtlCapacityVolume (number): Maximum truck capacity regarding the volume dimension for outbound transports.
        - minimumFtlCosts (number): Minimum costs of a full truck.
        - outboundFtlCost1DistanceUnit (number): Costs per km/mi for the first distance unit.
        - outboundFtlCost1000DistanceUnit (number): Costs per km/mi for the 1000. distance unit.
        - minimumLtlCosts (number): Minimum costs of an LTL truck.
        - outboundCostsHalfFtlPercent (number): Costs for 1/2 truck.
        - minimumInboundReplenishmentFrequency (number): Minimum number of shipments that should go between a factory and a warehouse.
        - minimumOutboundReplenishmentFrequency (number): Minimum number of shipments that should go between a warehouse and a customer.
        - weightCostFunction (str): ID of the relevant weight cost function.
        - volumeCostFunction (str): ID of the relevant volume cost function.
    customers (pd.DataFrame): Customer data with its address and its shipment structure for the optimization run.
        - name (str): Customer name.
        - country (str): Country code or country name.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        - weight (number): Average weight of a shipment that goes to this customer.
        - volume (number): Average volume of a shipment that goes to this customer.
        - numberOfShipments (number): Total number of shipments (integer value) the customer receives.
        - productSegments (str): Influences which warehouses could be used to fulfill the customer demand.
        - factory (str): The factory that delivers to the customer.
        - warehouse (str): The warehouse that delivers to the customer.
        - maximumWarehouseDistance (number): Maximum allowed distance from the assigned warehouse.
        - penaltyCostsWarehouseDistance (number): Penalty costs per distance unit.
        - customerGroup (str): A grouping field used to route multiple orders through the same warehouse.
    product_segments (pd.DataFrame): Data for the existing product segments, and the available warehouses for each of segment. 
        - segmentName (str): Name of a specific product segment.
        - availableWarehouses (str): List of warehouses which could be used for a specific segment separated with a semicolon.
    transport_costs (pd.DataFrame): Defining detailed transport costs for all factory-warehouse and warehouse-customer relations.
        - startLocationName (str): Start name of the relation.
        - endLocationName (str): End name of the relation.
        - weightCapacity (number): Capacity for the weight dimension for the transport unit.
        - volumeCapacity (number): Capacity for the volume dimension for the transport unit.
        - costsPerUnit (number): Costs per transport unit.
    transport_costs_rules (pd.DataFrame): Defining general rules and the corresponding vehicle capacities transport costs.
        - fromCountryIso2 (str): ISO 2 code of the sender country.
        - fromZip (str): Postal code of the sender.
        - fromName (str): Start name of the relation. 
        - toCountryIso2 (str): ISO 2 code of the recipient country.
        - toZip (str): Postal code of the recipient.
        - toName (str): End name of the relation. 
        - layer (str): The direction of the flow, can only be "inbound" (from factory to warehouse) or "outbound" 
                       (from warehouse to customer).
        - distance (number): Distance between sender and recipient.
        - weightCapacityPerUnit (number): Capacity for the weight dimension for the transport unit.
        - volumeCapacityPerUnit (number): Capacity for the volume dimension for the transport unit.
        - costsPerUnit (number): Costs per transport unit.
    stepwise_function_weight (pd.DataFrame): Data for a stepwise (piecewise) cost function for the weight dimension.   
        - stepFunctionId (str): ID of the respective cost function.
        - stepEnd (number): An upper limit for a specific step.
        - fixedCost (number): Fixed cost for a specific step.
        - costPerWeightUnit (number): Cost per weight unit for a specific step.
    stepwise_function_volume (pd.DataFrame): Data for a stepwise (piecewise) cost function for the volume dimension.   
        - stepFunctionId (str): ID of the respective cost function.
        - stepEnd (number): An upper limit for a specific step.
        - fixedCost (number): Fixed cost for a specific step.
        - costPerVolumeUnit (number): Cost per volume unit for a specific step.
    distance_limits (pd.DataFrame): Data specifying the maximum distance within which a given percentage of total demand must be met.
        - distanceLimit (number):  Maximum possible distance between the warehouse and delivery points.
        - weightPercentage (number): Minimum percentage of total weight that needs to be covered within the specified distance.
        - volumePercentage (number): Minimum percentage of total volume that needs to be covered within the specified distance.
        - distanceLimitPenalty (number): Penalty cost, if the constraint would not be fulfilled.

    parametrs (dict): A dictionary containing parameters distanceUnit (enum 'km' or 'mi'), vehicleType (enum 'car' or 
                      'truck'), streetLevel (boolean), inboundConsolidation (boolean), outboundConsolidation (boolean), 
                      allowMultisourcing (boolean), minWarehouses (integer number), maxWarehouses (integer number).

    api_key (str): The Log-hub API key for accessing the network design service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                            'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                            'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]: Four dataframes with the information about opened warehouses, factory and customer assignment, and solution kpis. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    factories_mandatory_columns = {
       'name': 'str', 'country': 'str', 'outboundFtlCapacityWeight': 'float', 'outboundFtlCapacityVolume': 'float', 'outboundFtlCost1DistanceUnit': 'float', 'outboundFtlCost1000DistanceUnit': 'float', 'outboundCostsHalfFtlPercent': 'float'
    }
    factories_optional_columns = { 'id': 'float', 'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str'}
    factories_optional_float = ['minimumFtlCosts', 'minimumLtlCosts']
    warehouses_mandatory_columns = {
        'name': 'str', 'country': 'str', 'fixed': 'float', 'minWeight': 'float', 'maxWeight': 'float', 'minVolume': 'float', 'maxVolume': 'float','fixedCosts': 'float','costsPerWeightUnit': 'float', 'costsPerVolumeUnit': 'float', 'outboundFtlCapacityWeight': 'float', 'outboundFtlCapacityVolume': 'float', 'outboundFtlCost1DistanceUnit': 'float', 'outboundFtlCost1000DistanceUnit': 'float', 'outboundCostsHalfFtlPercent': 'float'
    }
    warehouses_optional_columns = {'id': 'float', 'state' : 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str', 'weightCostFunction': 'str', 'volumeCostFunction': 'str'}
    warehouses_optional_float = ['penaltyCostsWeight', 'penaltyCostsVolume', 'minimumFtlCosts', 'minimumLtlCosts', 'minimumInboundReplenishmentFrequency', 'minimumOutboundReplenishmentFrequency']
    customers_mandatory_columns = {
        'name': 'str', 'country': 'str', 'weight': 'float', 'volume': 'float', 'numberOfShipments': 'float'
    }
    customers_optional_columns = {'id': 'float', 'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str', 'productSegments': 'str', 'factory': 'str', 'warehouse': 'str', 'customerGroup': 'str'}
    customers_optional_float = ['maximumWarehouseDistance', 'penaltyCostsWarehouseDistance']
    product_segments_optional_columns = {
        'id': 'float', 'segmentName': 'str', 'availableWarehouses': 'str'
    }
    transport_costs_mandatory_columns = {
        'startLocationName': 'str', 'endLocationName': 'str', 'volumeCapacity': 'float', 'costsPerUnit': 'float',
    }
    transport_costs_optional_columns = {'id': 'float'}
    transport_costs_optional_float = ['weightCapacity']
    transport_costs_rules_optional_columns = {
        'id': 'float', 'fromCountryIso2': 'str', 'fromZip': 'str',  'fromName': 'str', 'toCountryIso2': 'str', 'toZip': 'str', 'toName': 'str', 'layer': 'str'
    }
    transport_costs_rules_optional_float = ['distance', 'weightCapacityPerUnit', 'volumeCapacityPerUnit', 'costsPerUnit']
    stepwise_function_weight_optional_columns = {
        'id': 'float', 'stepFunctionId': 'str'
        }
    stepwise_function_weight_optional_float = ['stepEnd', 'fixedCost', 'costPerWeightUnit']
    stepwise_function_volume_optional_columns = {
        'id': 'float', 'stepFunctionId': 'str'
        }
    stepwise_function_volume_optional_float = ['stepEnd', 'fixedCost', 'costPerVolumeUnit']
    distance_limits_columns = {
        'id': 'float', 'distanceLimit': 'str',  'weightPercentage': 'str', 'volumePercentage': 'str', 'distanceLimitPenalty': 'str', 
        }

    # Validate and convert data types
    factories = validate_and_convert_data_types(factories, factories_mandatory_columns, 'mandatory', 'factories')
    if not factories is None:
        factories = validate_and_convert_data_types(factories, factories_optional_columns, 'optional', 'factories')
        if not factories is None:
            factories = convert_to_float(factories, factories_optional_float, 'optional')
            factories = convert_df_to_dict_excluding_nan(factories, factories_optional_float)

    warehouses = validate_and_convert_data_types(warehouses, warehouses_mandatory_columns, 'mandatory', 'warehouses')
    if not warehouses is None:
        warehouses = validate_and_convert_data_types(warehouses, warehouses_optional_columns, 'optional', 'warehouses')
        if not warehouses is None:
            warehouses = convert_to_float(warehouses, warehouses_optional_float, 'optional')
            warehouses = convert_df_to_dict_excluding_nan(warehouses,warehouses_optional_float)

    customers = validate_and_convert_data_types(customers, customers_mandatory_columns, 'mandatory', 'customers')
    if not customers is None:
        customers = validate_and_convert_data_types(customers, customers_optional_columns, 'optional', 'customers')
        if not customers is None:
            customers = convert_to_float(customers, customers_optional_float, 'optional')
            customers = convert_df_to_dict_excluding_nan(customers, customers_optional_float)

    product_segments = validate_and_convert_data_types(product_segments, product_segments_optional_columns, 'optional', 'roduct segments')
    
    transport_costs = validate_and_convert_data_types(transport_costs, transport_costs_mandatory_columns, 'mandatory', 'transport costs')
    if not transport_costs is None:
        transport_costs = validate_and_convert_data_types(transport_costs, transport_costs_optional_columns, 'optional', 'transport costs')
        if not transport_costs is None:
            transport_costs = convert_to_float(transport_costs, transport_costs_optional_float, 'optional')
            transport_costs = convert_df_to_dict_excluding_nan(transport_costs, transport_costs_optional_float)

    transport_costs_rules = validate_and_convert_data_types(transport_costs_rules, transport_costs_rules_optional_columns, 'optional', 'transport costs rules')
    if not transport_costs_rules is None:
        transport_costs_rules = convert_to_float(transport_costs_rules, transport_costs_rules_optional_float, 'optional')
        transport_costs_rules = convert_df_to_dict_excluding_nan(transport_costs_rules, transport_costs_rules_optional_float)

    stepwise_function_weight = validate_and_convert_data_types(stepwise_function_weight, stepwise_function_weight_optional_columns, 'optional', 'stepwise function weight')
    if not stepwise_function_weight is None:
        stepwise_function_weight = convert_to_float(stepwise_function_weight, stepwise_function_weight_optional_float, 'optional')
        stepwise_function_weight = convert_df_to_dict_excluding_nan(stepwise_function_weight, stepwise_function_weight_optional_float)

    stepwise_function_volume = validate_and_convert_data_types(stepwise_function_volume, stepwise_function_volume_optional_columns, 'optional', 'stepwise function volume')
    if not stepwise_function_volume is None:
        stepwise_function_volume = convert_to_float(stepwise_function_volume, stepwise_function_volume_optional_float, 'optional')
        stepwise_function_volume = convert_df_to_dict_excluding_nan(stepwise_function_volume, stepwise_function_volume_optional_float)

    distance_limits = validate_and_convert_data_types(distance_limits, distance_limits_columns, 'optional', 'distance limits')

    if any(df is None for df in [factories, warehouses, customers, product_segments, transport_costs, transport_costs_rules, stepwise_function_weight, stepwise_function_volume, distance_limits]):
        return None
    
    url = create_url("networkdesignpluslongrun")
    
    headers = create_headers(api_key)

    payload = {
        'factories': factories,
        'warehouses': warehouses,
        'customers': customers,
        'productSegments': product_segments.to_dict(orient='records'),
        'transportCosts': transport_costs,
        'transportsCostsRules': transport_costs_rules,
        'stepwiseCostFunctionWeight': stepwise_function_weight,
        'stepwiseCostFunctionVolume': stepwise_function_volume,
        'distanceLimits': distance_limits.to_dict(orient='records'),
        'parameters': parameters
    }
    payload = save_scenario_check(save_scenario, payload)

    response_data = post_method(url, payload, headers, "network design plus")
    if response_data is None:
        return None
    else:
        result = response_data['result']
        get_method_result = get_method(result['apiServer'], result['url'], {"authorization": f"apikey {api_key}"}, "network design plus")
        if get_method_result is None:
            return None
        else:
            open_warehouses = pd.DataFrame(get_method_result['openWarehouses'])
            factory_assignment = pd.DataFrame(get_method_result['factoryAssignment'])
            customer_assignment = pd.DataFrame(get_method_result['customerAssignment'])
            solution_kpis = pd.DataFrame(get_method_result['solutionKpis'])
            if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
                create_buttons()
            if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
                logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
            return open_warehouses, factory_assignment, customer_assignment,solution_kpis

def forward_network_design_plus_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'NetworkDesignPlusAddresses.xlsx')
    factories_df = pd.read_excel(data_path, sheet_name='factories', usecols='A:N').fillna("")
    warehouses_df = pd.read_excel(data_path, sheet_name='warehouses', usecols='A:AB').fillna("")
    customers_df = pd.read_excel(data_path, sheet_name='customers', usecols='A:P').fillna("")
    product_segments_df = pd.read_excel(data_path, sheet_name='productSegments', usecols='A:C').fillna("")
    transport_costs_df = pd.read_excel(data_path, sheet_name='transportCosts', usecols='A:F').fillna("")
    transport_costs_rules_df = pd.read_excel(data_path, sheet_name='transportCostsRules', usecols='A:L').fillna("")
    stepwise_function_weight_df = pd.read_excel(data_path, sheet_name='stepwiseCostFunctionWeight', usecols='A:E').fillna("")
    stepwise_function_volume_df = pd.read_excel(data_path, sheet_name='stepwiseCostFunctionVolume', usecols='A:E').fillna("")
    distance_limits_df = pd.read_excel(data_path, sheet_name='distanceLimits', usecols='A:E').fillna("")
    parameters = {
        "distanceUnit": "km",
        "vehicleType": "car",
        "streetLevel": False,
        "inboundConsolidation": True,
        "outboundConsolidation": False,
        "allowMultisourcing": True,
        "minWarehouses": 1,
        "maxWarehouses": 2

    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'factories': factories_df, 'warehouses': warehouses_df, 'customers': customers_df, 'productSegments': product_segments_df, 'transportCosts': transport_costs_df, 'transportCostsRules': transport_costs_rules_df, 'stepwiseCostFunctionWeight': stepwise_function_weight_df, 'stepwiseCostFunctionVolume': stepwise_function_volume_df, 'distanceLimits': distance_limits_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}

def reverse_network_design_plus(factories: pd.DataFrame, warehouses: pd.DataFrame, customers: pd.DataFrame, product_segments: pd.DataFrame, transport_costs: pd.DataFrame, transport_costs_rules: pd.DataFrame, stepwise_function_weight: pd.DataFrame, stepwise_function_volume: pd.DataFrame, distance_limits: pd.DataFrame, parameters: dict, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]]:
    """
    Perform reverse network design plus based on factories, warehouses, customers, product segments, transport costs, transport costs rules, stepwise functions for weight and volume, and distance limits.

    This function takes DataFrames of factories, warehouses, customers, product segments, transport costs, transport costs rules, stepwise functions for weight and volume, distance limits and an API key, and performs reverse network design plus using the Log-hub service. 

    Parameters:
    factories (pd.DataFrame): A pandas DataFrame containing the factories information for the optimization run.
        Required columns and their types:
        - name (str): Factory name.
        - latitude (number): Factory latitude.
        - longitude (number): Factory longitude.
        - outboundFtlCapacityWeight (number): Maximum truck capacity regarding the weight dimension for outbound transports.
        - outboundFtlCapacityVolume (number): Maximum truck capacity regarding the volume dimension for outbound transports.
        - minimumFtlCosts (number): Minimum costs of a full truck.
        - outboundFtlCost1DistanceUnit (number): Costs per km/mi for the first distance unit.
        - outboundFtlCost1000DistanceUnit (number): Costs per km/mi for the 1000. distance unit.
        - minimumLtlCosts (number): Minimum costs of an LTL truck.
        - outboundCostsHalfFtlPercent (number): Costs for 1/2 truck.
    warehouses (pd.DataFrame): A pandas DataFrame containing the warehouses information for the optimization run.
        - name (str): Warehouse name.
        - latitude (number): Warehouse latitude.
        - longitude (number): Warehouse longitude.
        - fixed (0 or 1): If the warehouse should be in every case part of the result, then mark it as fixed (1).
        - minWeight (number): The weight that should be at least assigned to this warehouse.
        - maxWeight (number): The maximum weight capacity of the warehouse.
        - penaltyCostsWeight (number): Penalty costs per weight unit. 
        - minVolume (number): The volume that should be at least assigned to this warehouse.
        - maxVolume (number): The maximum volume capacity of the warehouse.
        - penaltyCostsVolume (number): Penalty costs per volume unit. 
        - fixedCosts (number):  Fixed warehouse costs.
        - costsPerWeightUnit (number): Costs by weight unit.
        - costsPerVolumeUnit (number): Costs by volume unit.
        - outboundFtlCapacityWeight (number): Maximum truck capacity regarding the weight dimension for outbound transports.
        - outboundFtlCapacityVolume (number): Maximum truck capacity regarding the volume dimension for outbound transports.
        - minimumFtlCosts (number): Minimum costs of a full truck.
        - outboundFtlCost1DistanceUnit (number): Costs per km/mi for the first distance unit.
        - outboundFtlCost1000DistanceUnit (number): Costs per km/mi for the 1000. distance unit.
        - minimumLtlCosts (number): Minimum costs of an LTL truck.
        - outboundCostsHalfFtlPercent (number): Costs for 1/2 truck.
        - minimumInboundReplenishmentFrequency (number): Minimum number of shipments that should go between a factory and a warehouse.
        - minimumOutboundReplenishmentFrequency (number): Minimum number of shipments that should go between a warehouse and a customer.
        - weightCostFunction (str): ID of the relevant weight cost function.
        - volumeCostFunction (str): ID of the relevant volume cost function.
    customers (pd.DataFrame): Customer data with its address and its shipment structure for the optimization run.
        - name (str): Customer name.
        - latitude (number): Customer latitude.
        - longitude (number): Customer longitude.
        - weight (number): Average weight of a shipment that goes to this customer.
        - volume (number): Average volume of a shipment that goes to this customer.
        - numberOfShipments (number): Total number of shipments (integer value) the customer receives.
        - productSegments (str): Influences which warehouses could be used to fulfill the customer demand.
        - factory (str): The factory that delivers to the customer.
        - warehouse (str): The warehouse that delivers to the customer.
        - maximumWarehouseDistance (number): Maximum allowed distance from the assigned warehouse.
        - penaltyCostsWarehouseDistance (number): Penalty costs per distance unit.
        - customerGroup (str): A grouping field used to route multiple orders through the same warehouse.
    product_segments (pd.DataFrame): Data for the existing product segments, and the available warehouses for each of segment. 
        - segmentName (str): Name of a specific product segment.
        - availableWarehouses (str): List of warehouses which could be used for a specific segment separated with a semicolon.
    transport_costs (pd.DataFrame): Defining detailed transport costs for all factory-warehouse and warehouse-customer relations.
        - startLocationName (str): Start name of the relation.
        - endLocationName (str): End name of the relation.
        - weightCapacity (number): Capacity for the weight dimension for the transport unit.
        - volumeCapacity (number): Capacity for the volume dimension for the transport unit.
        - costsPerUnit (number): Costs per transport unit.
    transport_costs_rules (pd.DataFrame): Defining general rules and the corresponding vehicle capacities transport costs.
        - fromCountryIso2 (str): ISO 2 code of the sender country.
        - fromZip (str): Postal code of the sender.
        - fromName (str): Start name of the relation. 
        - toCountryIso2 (str): ISO 2 code of the recipient country.
        - toZip (str): Postal code of the recipient.
        - toName (str): End name of the relation. 
        - layer (str): The direction of the flow, can only be "inbound" (from factory to warehouse) or "outbound" 
                       (from warehouse to customer).
        - distance (number): Distance between sender and recipient.
        - weightCapacityPerUnit (number): Capacity for the weight dimension for the transport unit.
        - volumeCapacityPerUnit (number): Capacity for the volume dimension for the transport unit.
        - costsPerUnit (number): Costs per transport unit.
    stepwise_function_weight (pd.DataFrame): Data for a stepwise (piecewise) cost function for the weight dimension.   
        - stepFunctionId (str): ID of the respective cost function.
        - stepEnd (number): An upper limit for a specific step.
        - fixedCost (number): Fixed cost for a specific step.
        - costPerWeightUnit (number): Cost per weight unit for a specific step.
    stepwise_function_volume (pd.DataFrame): Data for a stepwise (piecewise) cost function for the volume dimension.   
        - stepFunctionId (str): ID of the respective cost function.
        - stepEnd (number): An upper limit for a specific step.
        - fixedCost (number): Fixed cost for a specific step.
        - costPerVolumeUnit (number): Cost per volume unit for a specific step.
    distance_limits (pd.DataFrame): Data specifying the maximum distance within which a given percentage of total demand must be met.
        - distanceLimit (number):  Maximum possible distance between the warehouse and delivery points.
        - weightPercentage (number): Minimum percentage of total weight that needs to be covered within the specified distance.
        - volumePercentage (number): Minimum percentage of total volume that needs to be covered within the specified distance.
        - distanceLimitPenalty (number): Penalty cost, if the constraint would not be fulfilled.

    parametrs (dict): A dictionary containing parameters distanceUnit (enum 'km' or 'mi'), vehicleType (enum 'car' or 
                      'truck'), streetLevel (boolean), inboundConsolidation (boolean), outboundConsolidation (boolean), 
                      allowMultisourcing (boolean), minWarehouses (integer number), maxWarehouses (integer number).

    api_key (str): The Log-hub API key for accessing the network design service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                            'saveScenario' (boolean), 'overwriteScenario' (boolean), 'workspaceId' (str) and
                            'scenarioName' (str).
    
    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]: Four dataframes with the information about opened warehouses, factory and customer assignment, and solution kpis. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['dashboard'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📊 Open Dashboard", "📋 Show Input Dataset", "📋 Show Output Dataset"])

    factories_mandatory_columns = {
        'name': 'str', 'latitude': 'float', 'longitude': 'float', 'outboundFtlCapacityWeight': 'float', 'outboundFtlCapacityVolume': 'float', 'outboundFtlCost1DistanceUnit': 'float', 'outboundFtlCost1000DistanceUnit': 'float', 'outboundCostsHalfFtlPercent': 'float'
    }
    factories_optional_float = ['id', 'minimumFtlCosts', 'minimumLtlCosts']
    warehouses_mandatory_columns = {
        'name': 'str', 'latitude': 'float', 'longitude': 'float', 'fixed': 'float', 'minWeight': 'float', 'maxWeight': 'float', 'minVolume': 'float', 'maxVolume': 'float', 'fixedCosts': 'float','costsPerWeightUnit': 'float', 'costsPerVolumeUnit': 'float', 'outboundFtlCapacityWeight': 'float', 'outboundFtlCapacityVolume': 'float', 'outboundFtlCost1DistanceUnit': 'float', 'outboundFtlCost1000DistanceUnit': 'float', 'outboundCostsHalfFtlPercent': 'float'
    }
    warehouses_optional_columns = {'weightCostFunction': 'str', 'volumeCostFunction': 'str'}
    warehouses_optional_float = ['id', 'penaltyCostsWeight', 'penaltyCostsVolume', 'minimumFtlCosts', 'minimumLtlCosts', 'minimumInboundReplenishmentFrequency', 'minimumOutboundReplenishmentFrequency']
    customers_mandatory_columns = {
        'name': 'str', 'latitude': 'float', 'longitude': 'float', 'weight': 'float', 'volume': 'float', 'numberOfShipments': 'float'
    }
    customers_optional_columns = {'productSegments': 'str', 'factory': 'str', 'warehouse': 'str', 'customerGroup': 'str'}
    customers_optional_float = ['id', 'maximumWarehouseDistance', 'penaltyCostsWarehouseDistance']
    product_segments_optional_columns = {
        'id': 'float', 'segmentName': 'str', 'availableWarehouses': 'str'
    }
    transport_costs_mandatory_columns = {
        'startLocationName': 'str', 'endLocationName': 'str', 'volumeCapacity': 'float', 'costsPerUnit': 'float',
    }
    transport_costs_optional_columns = {'id': 'float'}
    transport_costs_optional_float = ['weightCapacity']
    transport_costs_rules_optional_columns = {
        'id': 'float', 'fromCountryIso2': 'str', 'fromZip': 'str',  'fromName': 'str', 'toCountryIso2': 'str', 'toZip': 'str', 'toName': 'str', 'layer': 'str'
    }
    transport_costs_rules_optional_float = ['distance', 'weightCapacityPerUnit', 'volumeCapacityPerUnit', 'costsPerUnit']
    stepwise_function_weight_optional_columns = {
        'id': 'float', 'stepFunctionId': 'str'
        }
    stepwise_function_weight_optional_float = ['stepEnd', 'fixedCost', 'costPerWeightUnit']
    stepwise_function_volume_optional_columns = {
        'id': 'float', 'stepFunctionId': 'str'
        }
    stepwise_function_volume_optional_float = ['stepEnd', 'fixedCost', 'costPerVolumeUnit']
    distance_limits_columns = {
        'id': 'float', 'distanceLimit': 'str',  'weightPercentage': 'str', 'volumePercentage': 'str', 'distanceLimitPenalty': 'str', 
        }

    # Validate and convert data types
    factories = validate_and_convert_data_types(factories, factories_mandatory_columns, 'mandatory', 'factories')
    if not factories is None:
        factories = convert_to_float(factories, factories_optional_float, 'optional')
        factories = convert_df_to_dict_excluding_nan(factories, factories_optional_float)

    warehouses = validate_and_convert_data_types(warehouses, warehouses_mandatory_columns, 'mandatory', 'warehouses')
    if not warehouses is None:
        warehouses = validate_and_convert_data_types(warehouses, warehouses_optional_columns, 'optional', 'warehouses')
        if not warehouses is None:
            warehouses = convert_to_float(warehouses, warehouses_optional_float, 'optional')
            warehouses = convert_df_to_dict_excluding_nan(warehouses,warehouses_optional_float)

    customers = validate_and_convert_data_types(customers, customers_mandatory_columns, 'mandatory', 'customers')
    if not customers is None:
        customers = validate_and_convert_data_types(customers, customers_optional_columns, 'optional', 'customers')
        if not customers is None:
            customers = convert_to_float(customers, customers_optional_float, 'optional')
            customers = convert_df_to_dict_excluding_nan(customers, customers_optional_float)

    product_segments = validate_and_convert_data_types(product_segments, product_segments_optional_columns, 'optional', 'product segments')
    
    transport_costs = validate_and_convert_data_types(transport_costs, transport_costs_mandatory_columns, 'mandatory', 'transport costs')
    if not transport_costs is None:
        transport_costs = validate_and_convert_data_types(transport_costs, transport_costs_optional_columns, 'optional', 'transport costs')
        if not transport_costs is None:
            transport_costs = convert_to_float(transport_costs, transport_costs_optional_float, 'optional')
            transport_costs = convert_df_to_dict_excluding_nan(transport_costs, transport_costs_optional_float)

    transport_costs_rules = validate_and_convert_data_types(transport_costs_rules, transport_costs_rules_optional_columns, 'optional', 'transport costs rules')
    if not transport_costs_rules is None:
        transport_costs_rules = convert_to_float(transport_costs_rules, transport_costs_rules_optional_float, 'optional')
        transport_costs_rules = convert_df_to_dict_excluding_nan(transport_costs_rules, transport_costs_rules_optional_float)
    
    stepwise_function_weight = validate_and_convert_data_types(stepwise_function_weight, stepwise_function_weight_optional_columns, 'optional', 'stepwise function weight')
    if not stepwise_function_weight is None:
        stepwise_function_weight = convert_to_float(stepwise_function_weight, stepwise_function_weight_optional_float, 'optional')
        stepwise_function_weight = convert_df_to_dict_excluding_nan(stepwise_function_weight, stepwise_function_weight_optional_float)

    stepwise_function_volume = validate_and_convert_data_types(stepwise_function_volume, stepwise_function_volume_optional_columns, 'optional', 'stepwise function volume')
    if not stepwise_function_volume is None:
        stepwise_function_volume = convert_to_float(stepwise_function_volume, stepwise_function_volume_optional_float, 'optional')
        stepwise_function_volume = convert_df_to_dict_excluding_nan(stepwise_function_volume, stepwise_function_volume_optional_float)

    distance_limits = validate_and_convert_data_types(distance_limits, distance_limits_columns, 'optional', 'distance limits')

    if any(df is None for df in [factories, warehouses, customers, product_segments, transport_costs, transport_costs_rules, stepwise_function_weight, stepwise_function_volume, distance_limits]):
        return None
    
    url = create_url("reversenetworkdesignpluslongrun")
    
    headers = create_headers(api_key)

    payload = {
        'factories': factories,
        'warehouses': warehouses,
        'customers': customers,
        'productSegments': product_segments.to_dict(orient='records'),
        'transportCosts': transport_costs,
        'transportsCostsRules': transport_costs_rules,
        'stepwiseCostFunctionWeight': stepwise_function_weight,
        'stepwiseCostFunctionVolume': stepwise_function_volume,
        'distanceLimits': distance_limits.to_dict(orient='records'),
        'parameters': parameters
    }
    payload = save_scenario_check(save_scenario, payload)

    response_data = post_method(url, payload, headers, "network design plus")
    if response_data is None:
        return None
    else:
        result = response_data['result']
        get_method_result = get_method(result['apiServer'], result['url'], {"authorization": f"apikey {api_key}"}, "network design plus")
        if get_method_result is None:
            return None
        else:
            open_warehouses = pd.DataFrame(get_method_result['openWarehouses'])
            factory_assignment = pd.DataFrame(get_method_result['factoryAssignment'])
            customer_assignment = pd.DataFrame(get_method_result['customerAssignment'])
            solution_kpis = pd.DataFrame(get_method_result['solutionKpis'])
            if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
                create_buttons()
            if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
                logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
            return open_warehouses, factory_assignment, customer_assignment,solution_kpis
        
def reverse_network_design_plus_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'NetworkDesignPlusReverse.xlsx')
    factories_df = pd.read_excel(data_path, sheet_name='factories', usecols='A:K').fillna("")
    warehouses_df = pd.read_excel(data_path, sheet_name='warehouses', usecols='A:Y').fillna("")
    customers_df = pd.read_excel(data_path, sheet_name='customers', usecols='A:M').fillna("")
    product_segments_df = pd.read_excel(data_path, sheet_name='productSegments', usecols='A:C').fillna("")
    transport_costs_df = pd.read_excel(data_path, sheet_name='transportCosts', usecols='A:F').fillna("")
    transport_costs_rules_df = pd.read_excel(data_path, sheet_name='transportCostsRules', usecols='A:L').fillna("")
    stepwise_function_weight_df = pd.read_excel(data_path, sheet_name='stepwiseCostFunctionWeight', usecols='A:E').fillna("")
    stepwise_function_volume_df = pd.read_excel(data_path, sheet_name='stepwiseCostFunctionVolume', usecols='A:E').fillna("")
    distance_limits_df = pd.read_excel(data_path, sheet_name='distanceLimits', usecols='A:E').fillna("")
    parameters = {
        "distanceUnit": "km",
        "vehicleType": "car",
        "streetLevel": False,
        "inboundConsolidation": True,
        "outboundConsolidation": False,
        "allowMultisourcing": True,
        "minWarehouses": 1,
        "maxWarehouses": 2

    }
    save_scenario = {
        'saveScenario': False,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        'scenarioName': 'Your scenario name'
    }
    return {'factories': factories_df, 'warehouses': warehouses_df, 'customers': customers_df, 'productSegments': product_segments_df, 'transportCosts': transport_costs_df, 'transportCostsRules': transport_costs_rules_df, 'stepwiseCostFunctionWeight': stepwise_function_weight_df, 'stepwiseCostFunctionVolume': stepwise_function_volume_df, 'distanceLimits': distance_limits_df, 'parameters': parameters, 'saveScenarioParameters': save_scenario}


import os
import pandas as pd
import warnings
from typing import Optional
import sys
import logging
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'pyloghub')))
from save_to_platform import save_scenario_check, create_button
from input_data_validation import validate_and_convert_data_types, convert_to_float, convert_df_to_dict_excluding_nan
from sending_requests import post_method, create_headers, create_url, get_workspace_entities

def forward_supply_chain_map_locations(addresses: pd.DataFrame, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[pd.DataFrame]:
    """
    Creates a map of locations based on the given addresses.

    This function takes a DataFrame of addresses with their layer, quantity and descriptions, along with an API key, 
    and creates a map of locations using the Log-hub Supply Chain Map service. 

    Parameters:
    addresses (pd.DataFrame): A pandas DataFrame containing addresses with their layer, quantity and descriptions.
        Each row should contain:
        - id (number): Identifier.
        - name (str): Location name.
        - country (str): Country code. Must have at least 1 character.
        - state (str): State code.
        - postalCode (str): Postal code.
        - city (str): City name.
        - street (str): Street name with house number.
        - layer (str): Visual settings of locations in the same layer can be adjusted together.
        - quantity (number): Numerical value affecting the size of the location symbol on the map.
        - nameDescription1, nameDescription2 (str): Additional location descriptions.

    api_key (str): The Log-hub API key for accessing the supply chain map service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'mergeWithExistingScenario (boolean), 'workspaceId' (str) and 'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    pd.DataFrame: A pandas DataFrame containg the addresses and their parsed latitude and longitude. Returns None if 
                  the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📋 Show Input Dataset", "📋 Show Output Dataset"])
    mandatory_columns = {
            'name': 'str', 'country': 'str'
        }
    optional_columns = {'state': 'str', 'postalCode': 'str', 'city': 'str', 'street': 'str', 'layer': 'str', 'nameDescription1': 'str', 'nameDescription2': 'str'}
    optional_floats = ['id', 'quantity']

    # Validate and convert data types
    addresses= validate_and_convert_data_types(addresses, mandatory_columns, 'mandatory', 'addresses')
    if not addresses is None:
        addresses = validate_and_convert_data_types(addresses, optional_columns, 'optional', 'addresses')
        if not addresses is None:
            addresses = convert_to_float(addresses, optional_floats, 'optional')
            addresses = convert_df_to_dict_excluding_nan(addresses, optional_floats)
    if addresses is None:
        return None

    url = create_url("supplychainmaplocations")
    
    headers = create_headers(api_key)
    payload = {
        "geocodingData": addresses,
    }
 
    payload = save_scenario_check(save_scenario, payload, "supplychainmaplocations")
    
    response_data = post_method(url, payload, headers, "supply chain map locations")

    if response_data is None:
        return None
    else:
        geocoding_result_df = pd.DataFrame(response_data['geocodingResult'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return geocoding_result_df

def forward_supply_chain_map_locations_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'MapLocationsAddresses.xlsx')
    addresses_df = pd.read_excel(data_path, sheet_name='addresses', usecols='A:K', dtype={'postalCode': str})

    save_scenario = {
        'saveScenario': True,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        "mergeWithExistingScenario": False,
        'scenarioName': 'Your scenario name'
    }
    return {'addresses': addresses_df, 'saveScenarioParameters': save_scenario}

def reverse_supply_chain_map_locations(coordinates: pd.DataFrame, api_key: str, save_scenario = {}, show_buttons = False) -> Optional[pd.DataFrame]:
    """
    Creates a map of locations based on the given coordinates.

    This function takes a DataFrame of coordinates with their layer, quantity and descriptions, along with an API key, 
    and creates a map of locations using the Log-hub Supply Chain Map service. 

    Parameters:
    coordinates (pd.DataFrame): A pandas DataFrame containing coordinates with their layer, quantity and descriptions.
        Each row should contain:
        - id (number): Identifier.
        - name (str): Location name.
        - latitude (number): Latitude of the location.
        - longitude (number): Longitude of the location.
        - layer (str): Visual settings of locations in the same layer can be adjusted together.
        - quantity (number): Numerical value affecting the size of the location symbol on the map.
        - nameDescription1, nameDescription2 (str): Additional location descriptions.

    api_key (str): The Log-hub API key for accessing the supply chain map service.

    save_scenario (dict): A dictionary containg information about saving scenario, empty by default. Allowed key vales are
                        'saveScenario' (boolean), 'overwriteScenario' (boolean), 'mergeWithExistingScenario (boolean), 'workspaceId' (str) and 'scenarioName' (str).

    show_buttons (boolean): If this parameter is set to True and the scenario is saved on the platform, the buttons linking to the output results, map, dashboard and the input table 
                           will be created. If the scenario is not saved, a proper message will be shown.

    Returns:
    pd.DataFrame: A pandas DataFrame containg the coordinates. Returns None if the process fails.
    """
    def create_buttons():
        links = get_workspace_entities(save_scenario, api_key)
        create_button(links = [links['map'], links['inputDataset'], links['outputDataset']], texts = ["🌍 Open Map", "📋 Show Input Dataset", "📋 Show Output Dataset"])
    mandatory_columns = {'latitude':'float', 'longitude': 'float',}
    optional_columns = {'name': 'str', 'layer': 'str', 'nameDescription1': 'str', 'nameDescription2': 'str'}
    optional_floats = ['id', 'quantity']

    # Validate and convert data types
    coordinates = validate_and_convert_data_types(coordinates, mandatory_columns, 'mandatory', 'coordinates')
    if not coordinates is None:
        coordinates = validate_and_convert_data_types(coordinates, optional_columns, 'optional', 'coordinates')
        if not coordinates is None:
            coordinates = convert_to_float(coordinates, optional_floats, 'optional')
            coordinates = convert_df_to_dict_excluding_nan(coordinates, optional_floats)
    if coordinates is None:
        return None

    url = create_url("reversesupplychainmaplocations")
    
    headers = create_headers(api_key)
    payload = {
        "geocodingData": coordinates,
    }
    
    payload = save_scenario_check(save_scenario, payload, "reversesupplychainmaplocations")
    response_data = post_method(url, payload, headers, "reverse supply chain map locations")
    if response_data is None:
        return None
    else:
        geocoding_data_df = pd.DataFrame(response_data['geocodingData'])
        if (show_buttons and payload['saveScenarioParameters']['saveScenario']):
            create_buttons()
        if (not payload['saveScenarioParameters']['saveScenario'] and show_buttons):
            logging.info("Please, save the scenario in order to create the buttons for opening the results on the platform.")
        return geocoding_data_df

def reverse_supply_chain_map_locations_sample_data():
    warnings.simplefilter("ignore", category=UserWarning)
    data_path = os.path.join(os.path.dirname(__file__), 'sample_data', 'MapLocationsReverse.xlsx')
    coordinates_df = pd.read_excel(data_path, sheet_name='coordinates', usecols='A:H')

    save_scenario = {
        'saveScenario': True,
        'overwriteScenario': False,
        'workspaceId': 'Your workspace id',
        "mergeWithExistingScenario": False,
        'scenarioName': 'Your scenario name'
    }
    return {'coordinates': coordinates_df, 'saveScenarioParameters': save_scenario}